# Generated by Django 2.2.20 on 2021-04-21 18:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="invoiceconfiguration",
            name="separate_area_access_charges",
            field=models.BooleanField(
                default=False, help_text="Check this box to display area access charges as a separate invoice category"
            ),
        ),
        migrations.AddField(
            model_name="invoiceconfiguration",
            name="separate_tool_usage_charges",
            field=models.BooleanField(
                default=False, help_text="Check this box to display tool usage charges as a separate invoice category"
            ),
        ),
        migrations.AddField(
            model_name="invoiceconfiguration",
            name="separate_consumable_charges",
            field=models.BooleanField(
                default=False, help_text="Check this box to display supplies/material as a separate invoice category"
            ),
        ),
        migrations.AddField(
            model_name="invoiceconfiguration",
            name="separate_custom_charges",
            field=models.BooleanField(
                default=False, help_text="Check this box to display custom charges as a separate invoice category"
            ),
        ),
        migrations.AddField(
            model_name="invoiceconfiguration",
            name="separate_missed_reservation_charges",
            field=models.BooleanField(
                default=False,
                help_text="Check this box to display missed reservation charges as a separate invoice category",
            ),
        ),
        migrations.AddField(
            model_name="invoiceconfiguration",
            name="separate_staff_charges",
            field=models.BooleanField(
                default=False, help_text="Check this box to display staff charges as a separate invoice category"
            ),
        ),
        migrations.AddField(
            model_name="invoiceconfiguration",
            name="separate_training_charges",
            field=models.BooleanField(
                default=False, help_text="Check this box to display training charges as a separate invoice category"
            ),
        ),
        migrations.AddField(
            model_name="invoicesummaryitem",
            name="item_type",
            field=models.IntegerField(
                blank=True,
                choices=[
                    (1, "tool_usage"),
                    (2, "area_access"),
                    (3, "consumable"),
                    (4, "missed_reservation"),
                    (5, "staff_charge"),
                    (6, "training_session"),
                    (7, "custom_charge"),
                ],
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="invoicesummaryitem",
            name="summary_item_type",
            field=models.IntegerField(
                choices=[
                    (1, "item"),
                    (2, "sub_total"),
                    (3, "discount_sub_total"),
                    (4, "discount"),
                    (5, "tax"),
                    (6, "other"),
                    (7, "fund"),
                ]
            ),
        ),
    ]



import copy
import importlib
from collections import defaultdict
from copy import copy
from datetime import date, datetime, time, timedelta, timezone
from decimal import Decimal
from functools import partial
from typing import Dict, Iterable, List, Optional, Union

from NEMO.models import (
    Account,
    AreaAccessRecord,
    ConsumableWithdraw,
    Project,
    Reservation,
    StaffCharge,
    TrainingSession,
    UsageEvent,
    User,
)
from NEMO.typing import QuerySetType
from NEMO.utilities import format_daterange, format_datetime, get_month_timeframe
from django.conf import settings
from django.contrib import messages
from django.contrib.admin.utils import get_deleted_objects
from django.db import transaction
from django.db.models import Q
from django.http import HttpRequest
from django.utils.timezone import localtime, make_aware, now

from NEMO_billing.invoices.customization import InvoiceCustomization
from NEMO_billing.invoices.exceptions import (
    InvoiceAlreadyExistException,
    InvoiceItemsNotInFacilityException,
    NoProjectCategorySetException,
    NoProjectDetailsSetException,
    NoRateSetException,
)
from NEMO_billing.invoices.models import (
    BillableItemType,
    Invoice,
    InvoiceConfiguration,
    InvoiceDetailItem,
    InvoiceSummaryItem,
    ProjectBillingDetails,
)
from NEMO_billing.invoices.utilities import display_amount, flatten, name_for_billable_item
from NEMO_billing.models import CoreFacility, CustomCharge
from NEMO_billing.rates.customization import BillingRatesCustomization
from NEMO_billing.rates.models import AreaHighestDailyRateGroup, Rate, RateType
from NEMO_billing.rates.utilities import get_pro_rated_minimum_charge_and_service_fee, get_rate_history
from NEMO_billing.templatetags.billing_tags import cap_discount_installed
from NEMO_billing.utilities import (
    get_billable_item_type_for_item,
    get_charges_amount_between,
    get_minutes_between_dates,
    round_decimal_amount,
)


# Helper class for invoices with separate detail items and summary items
class FullInvoice:
    def __init__(self, invoice: Invoice, details: List[InvoiceDetailItem], summaries: List[InvoiceSummaryItem]):
        self.invoice = invoice
        self.detail_items = details
        self.summary_items = summaries

    def save(self):
        if self.invoice:
            self.invoice.save_all(self.detail_items, self.summary_items)


class BillableItem(object):
    """Object representing a billable item (Tool usage, area access, consumable etc.)"""

    def __init__(self, item, project: Project, configuration: InvoiceConfiguration = None):
        self.item = item
        self.project: Optional[Project] = project
        self.configuration = configuration or InvoiceConfiguration()
        # Actual items
        self.tool = None
        self.area = None
        self.consumable = None
        # Rate (to be set later)
        self.rate: Optional[Rate] = None
        # Other properties
        self.start = getattr(item, "start", None) or getattr(item, "date", None)
        self.end = getattr(item, "end", None) or getattr(item, "date", None)
        self.user = getattr(item, "user", None) or getattr(item, "customer", None) or getattr(item, "trainee", None)
        self.validated: bool = getattr(item, "validated", False)
        self.validated_by: Optional[User] = getattr(item, "validated_by", None)
        self.waived: bool = getattr(item, "waived", False)
        self.waived_by: Optional[User] = getattr(item, "waived_by", None)
        self.waived_on: Optional[datetime] = getattr(item, "waived_on", None)
        self.proxy_user = (
            getattr(item, "operator", None)
            or getattr(item, "staff_member", None)
            or getattr(item, "creator", None)
            or getattr(item, "merchant", None)
            or getattr(item, "trainer", None)
            or getattr(getattr(item, "staff_charge", None), "staff_member", None)
        )
        # Setting tool, area, consumable and other properties
        self.item_type: Optional[BillableItemType] = None
        # Set quantity and amount if we know they cannot change later (Custom charge or supplies or training duration etc.)
        self.quantity: Optional[Decimal] = None
        self.amount: Optional[Decimal] = None
        self.core_facility: Optional[CoreFacility] = None
        self.rate_type: Optional[RateType] = None
        self.item_type = get_billable_item_type_for_item(item)
        if isinstance(item, UsageEvent):
            self.tool = item.tool
            self.rate_type = self.configuration.rates_types.get(RateType.Type.TOOL_USAGE)
        elif isinstance(item, AreaAccessRecord):
            self.area = item.area
            self.rate_type = self.configuration.rates_types.get(RateType.Type.AREA_USAGE)
        elif isinstance(item, StaffCharge):
            self.rate_type = self.configuration.rates_types.get(RateType.Type.STAFF_CHARGE)
            if not getattr(item, "core_facility", None):
                # We first check if there was an area access, in which case we will use the area's facility
                area_charge: Optional[AreaAccessRecord] = None
                if item.id:
                    area_charge = AreaAccessRecord.objects.filter(staff_charge_id=item.id).first()
                if area_charge:
                    self.core_facility = area_charge.area.core_facility
                else:
                    # Otherwise, check for tool usage on behalf of the same customer during the time
                    tool_usage: QuerySetType[UsageEvent] = UsageEvent.objects.filter(
                        operator=item.staff_member, user=item.customer, start__gt=item.start
                    )
                    if item.end:
                        tool_usage = tool_usage.filter(start__lte=item.end)
                    tool_usage: UsageEvent = tool_usage.first()
                    if tool_usage:
                        self.core_facility = tool_usage.tool.core_facility
        elif isinstance(item, Reservation):
            self.tool = item.tool
            self.area = item.area
            if self.tool:
                self.rate_type = self.configuration.rates_types.get(RateType.Type.TOOL_MISSED_RESERVATION)
            if self.area:
                self.rate_type = self.configuration.rates_types.get(RateType.Type.AREA_MISSED_RESERVATION)
        elif isinstance(item, CustomCharge):
            # Quantity and amount will never change, so set them now
            self.quantity = Decimal(1)
            self.amount = item.amount
        elif isinstance(item, TrainingSession):
            self.tool = item.tool
            # Training duration will never change
            self.quantity = Decimal(item.duration)
            if item.type == TrainingSession.Type.INDIVIDUAL:
                self.rate_type = self.configuration.rates_types.get(RateType.Type.TOOL_TRAINING_INDIVIDUAL)
            else:
                self.rate_type = self.configuration.rates_types.get(RateType.Type.TOOL_TRAINING_GROUP)
        elif isinstance(item, ConsumableWithdraw):
            self.consumable = item.consumable
            # Supply order quantity will never change
            self.quantity = Decimal(item.quantity)
            self.rate_type = self.configuration.rates_types.get(RateType.Type.CONSUMABLE)
        if not self.core_facility:
            self.core_facility = (
                getattr(item, "core_facility", None)
                or getattr(self.tool, "core_facility", None)
                or getattr(self.area, "core_facility", None)
                or getattr(self.consumable, "core_facility", None)
            )

    @property
    def name(self):
        return name_for_billable_item(self)

    @property
    def display_rate(self):
        return get_rate_with_currency(self.configuration, self.rate.display_rate()) if self.rate else None

    @property
    def display_amount(self):
        amount = display_amount(self.amount, self.configuration)
        return f"{amount}{' (waived)' if self.waived else ''}"

    @property
    def rate_time_name_add(self):
        return f" - {self.rate.time.name}" if self.rate and self.rate.time else ""

    @property
    def display_quantity(self):
        if self.item_type in [
            BillableItemType.CONSUMABLE,
            BillableItemType.CUSTOM_CHARGE,
            BillableItemType.MISSED_RESERVATION,
        ]:
            return self.quantity
        else:
            return timedelta(seconds=round(self.quantity * 60))

    def calculate_amount(self, rate: Rate, new_start, new_end, copy_original=True):
        new_billable = copy(self) if copy_original else self
        new_billable.rate = rate
        new_billable.start = new_start
        new_billable.end = new_end
        # Calculate quantity only if it wasn't already set
        if not new_billable.quantity:
            new_billable.quantity = get_minutes_between_dates(new_billable.start, new_billable.end)
        # Calculate amount only if not already set
        if not new_billable.amount:
            new_billable.amount = rate.calculate_amount(new_billable.quantity)
        return new_billable


def find_billable_rates(billable_item: BillableItem, configuration: InvoiceConfiguration) -> List[Rate]:
    return find_rates(
        billable_item.rate_type,
        billable_item.project,
        configuration,
        localtime(billable_item.end).date(),
        billable_item.tool,
        billable_item.area,
        billable_item.consumable,
    )


def find_rates(
    rate_type: RateType,
    project: Project,
    config: InvoiceConfiguration,
    as_of_date: date,
    tool=None,
    area=None,
    consumable=None,
) -> List[Rate]:
    category = None
    tool_id = None
    area_id = None
    consumable_id = None
    try:
        project.projectbillingdetails
    except (ProjectBillingDetails.DoesNotExist, AttributeError):
        # Attribute error just in case we don't have a project, which should never happen
        raise NoProjectDetailsSetException(project)
    if rate_type.category_specific:
        if config.rates_exist and not project.projectbillingdetails.category:
            raise NoProjectCategorySetException(rate_type, project)
        category = project.projectbillingdetails.category
    if rate_type.item_specific:
        if tool:
            tool_id = tool.id
        elif area:
            area_id = area.id
        elif consumable:
            consumable_id = consumable.id
    category_id = category.id if category else None
    matching_rates = get_rate_history(
        rate_type.id, config.all_rates, category_id, tool_id, area_id, consumable_id, as_of_date
    ).current_rate()
    base_rate = matching_rates.get(None, None)
    if not rate_type.can_have_rate_time():
        # We can only have one rate
        if base_rate:
            return [base_rate]
        else:
            raise NoRateSetException(rate_type, category, tool=tool, area=area, consumable=consumable)
    else:
        # We need at least one rate, and one of them being a base rate (without time)
        if not matching_rates or not base_rate:
            raise NoRateSetException(rate_type, category, tool=tool, area=area, consumable=consumable)
        return list(matching_rates.values())


def get_rate_with_currency(config: InvoiceConfiguration, rate: str):
    return f"{config.currency} {rate}" if config.currency else rate


class InvoiceDataProcessor(object):
    # Do all this in a transaction so everything gets rolled back if an error happened
    @transaction.atomic
    def generate_invoice_for_account(self, month, account, configuration, user, raise_if_exists=False) -> List[Invoice]:
        full_invoice_list = []
        if not InvoiceCustomization.get_bool("invoice_skip_inactive_accounts") or account.active:
            # Start and end will be included in the results. So for example for September data, use: 09/01 12am to 09/30 11:59:99
            start, end = get_month_timeframe(month)
            account_filter = Q(project__account_id=account.id)
            all_billable_items = self.get_billable_items(
                start, end, configuration, account_filter, account_filter, account_filter
            )
            projects = account.project_set.all()
            if InvoiceCustomization.get_bool("invoice_skip_inactive_projects"):
                projects = projects.filter(active=True)
            for project in projects:
                billables = [billable for billable in all_billable_items if billable.project == project]
                invoice = self.generate_invoice(start, end, project, configuration, user, billables, raise_if_exists)
                if invoice:
                    full_invoice_list.append(invoice)
            for full_invoice in full_invoice_list:
                full_invoice.save()
        return [full_invoice.invoice for full_invoice in full_invoice_list]

    def generate_invoice(
        self,
        start: datetime,
        end: datetime,
        project: Project,
        configuration: InvoiceConfiguration,
        user: User,
        billables: List[BillableItem],
        raise_if_exists=False,
    ) -> FullInvoice:
        try:
            project_details = ProjectBillingDetails.objects.get(project_id=project.id)
        except ProjectBillingDetails.DoesNotExist:
            raise NoProjectDetailsSetException(project)
        # If there is already an invoice for this project for those dates, don't regenerate it (unless void).
        existing_invoices = Invoice.objects.filter(
            start=start, end=end, project_details=project_details, voided_date=None
        )
        if existing_invoices.exists():
            if raise_if_exists:
                raise InvoiceAlreadyExistException(existing_invoices.first())
        # No existing invoices, continue
        elif not project_details.no_charge:
            # Project can be charged, so proceed
            return invoice_data_processor_class.process_data(
                start, end, project_details, configuration, user, billables
            )

    def process_data(
        self,
        start: datetime,
        end: datetime,
        project_details: ProjectBillingDetails,
        configuration: InvoiceConfiguration,
        user: User,
        billables: List[BillableItem],
    ) -> Optional[FullInvoice]:
        invoice = self.create_invoice(start, end, project_details, configuration, user)
        detail_items: List[InvoiceDetailItem] = list(
            map(partial(billable_to_invoice_detail_item, invoice=invoice), billables)
        )
        if settings.INVOICE_ALL_ITEMS_MUST_BE_IN_FACILITY:
            for detail_item in detail_items:
                if not detail_item.core_facility:
                    raise InvoiceItemsNotInFacilityException(detail_item)
        if detail_items:
            summary_items = self.get_invoice_summary_items(invoice, detail_items)
            return FullInvoice(invoice, detail_items, summary_items)

    def get_billable_items(
        self,
        start: datetime,
        end: datetime,
        config: InvoiceConfiguration,
        customer_filter=Q(),
        user_filter=Q(),
        trainee_filter=Q(),
        raise_no_rate=True,
    ) -> List[BillableItem]:
        items: List[BillableItem] = []
        config.init_rate_help()
        items.extend(self._area_access_records(start, end, config, customer_filter, raise_no_rate))
        items.extend(self._consumable_withdrawals(start, end, config, customer_filter, raise_no_rate))
        items.extend(self._missed_reservations(start, end, config, user_filter, raise_no_rate))
        items.extend(self._staff_charges(start, end, config, customer_filter, raise_no_rate))
        items.extend(self._training_sessions(start, end, config, trainee_filter, raise_no_rate))
        items.extend(self._tool_usages(start, end, config, user_filter, raise_no_rate))
        items.extend(self._custom_charges(start, end, config, customer_filter, raise_no_rate))
        self.process_daily_rates(items)
        return items

    def get_billable_items_with_charge_filters(
        self,
        start: datetime,
        end: datetime,
        config: InvoiceConfiguration,
        usage_filter,
        area_access_filter,
        staff_charges_filter,
        consumable_filter,
        reservation_filter,
        training_filter,
        custom_charges_filter,
        raise_no_rate=True,
    ) -> List[BillableItem]:
        items: List[BillableItem] = []
        config.init_rate_help()
        items.extend(self._area_access_records(start, end, config, area_access_filter, raise_no_rate))
        items.extend(self._consumable_withdrawals(start, end, config, consumable_filter, raise_no_rate))
        items.extend(self._missed_reservations(start, end, config, reservation_filter, raise_no_rate))
        items.extend(self._staff_charges(start, end, config, staff_charges_filter, raise_no_rate))
        items.extend(self._training_sessions(start, end, config, training_filter, raise_no_rate))
        items.extend(self._tool_usages(start, end, config, usage_filter, raise_no_rate))
        items.extend(self._custom_charges(start, end, config, custom_charges_filter, raise_no_rate))
        self.process_daily_rates(items)
        return items

    def process_daily_rates(self, items: List[BillableItem]):
        # This function will grab all the charges that have a daily rate and organize them by rate_id
        # Then go over all items with the same rate and keep track of which ones we've processed for each day
        # If we already processed one on the same day, set the following ones to 0
        daily_by_account = cap_discount_installed() and BillingRatesCustomization.get_bool("rates_daily_per_account")
        daily_rate_charges = defaultdict(list)
        for item in copy(items):
            if not item.waived and item.rate and item.rate.daily and item.rate.daily_split_multi_day_charges:
                # we need to split multi-day charges
                # to simplify things, we are deciding here to use the end date of the original charge as reference point
                # this means that if a charge spans 7 days, all 7 day charges will be added to the invoice that
                # includes the last day. They will not be added to previous invoices.
                items.extend(split_records_by_day(item))
                items.remove(item)
        for item in items:
            if not item.waived and item.rate and item.rate.daily:
                daily_rate_charges[str(item.rate.id)].append(item)
        group_area_charges_by_day: Dict[str, List[BillableItem]] = defaultdict(list)
        for daily_charges in daily_rate_charges.values():
            treated_day: Dict[str, List[datetime.date]] = defaultdict(list)
            for item in daily_charges:
                # prefixes here (a, p, u, g) are not exactly necessary but help with debugging to know the model type
                account_id_or_project_id = f"a{item.project.account.id}" if daily_by_account else f"p{item.project.id}"
                key = f"u{item.user.id}-{account_id_or_project_id}"
                # if the end is at midnight, it should count for the previous day, so let's subtract a microsecond to get that day
                # this will only have a real effect if it ends right at midnight, so it should be safe
                day: datetime.date = (item.end - timedelta(microseconds=1)).astimezone().date()
                # set up a list of daily area charges that are part of a group for later processing
                if item.item_type == BillableItemType.AREA_ACCESS:
                    area_group = AreaHighestDailyRateGroup.objects.filter(areas__in=[item.area]).first()
                    if area_group:
                        group_area_day_key = f"{key}-g{area_group.id}-d{day.strftime('%Y%m%d')}"
                        group_area_charges_by_day[group_area_day_key].append(item)
                if day in treated_day[key]:
                    # If we have already processed this day for this rate, set to 0
                    item.amount = 0
                else:
                    treated_day[key].append(day)
        # let's do a second pass to deal with zeroing out daily charges in the same group
        for group_area_daily_charges in group_area_charges_by_day.values():
            # first, sort grouped areas by most expensive (all daily rates at this point)
            sorted_group_area_daily_charges = sorted(group_area_daily_charges, key=lambda x: -x.rate.amount)
            for order, group_area_item in enumerate(sorted_group_area_daily_charges):
                # the first one is the most expensive, we can zero out all other ones
                if order != 0:
                    group_area_item.amount = 0

    def create_invoice(self, start, end, project_details, configuration, user: User) -> Invoice:
        invoice = Invoice()
        invoice.start = start
        invoice.end = end
        invoice.project_details = project_details
        invoice.configuration = configuration
        invoice.created_by = user
        invoice.total_amount = 0
        return invoice

    @transaction.atomic
    def void_invoice(self, invoice: Invoice, request: HttpRequest):
        # NOTE: we are voiding all invoices for this account,
        # in case daily rates were applied across projects of the same account
        invoice_account = invoice.project_details.project.account
        invoice_year, invoice_month = invoice.start.year, invoice.start.month
        for invoice in get_account_invoices(invoice_account, invoice_year, invoice_month):
            self.do_void_invoice(invoice, request)

    def do_void_invoice(self, invoice: Invoice, request: HttpRequest):
        if not invoice.voided_date:
            if hasattr(invoice.project_details.project, "projectprepaymentdetail"):
                invoice.project_details.project.projectprepaymentdetail.restore_funds(invoice, request)
            invoice.voided_date = now()
            invoice.voided_by = request.user
            invoice.save()
            invoice.send_voided_email()
            messages.success(
                request, f"Invoice {invoice.invoice_number} was successfully marked as void.", "data-speed=30000"
            )

    @transaction.atomic
    def delete_invoice(self, invoice: Invoice, request: HttpRequest):
        # NOTE: we are deleting all invoices for this account,
        # in case daily rates were applied across projects of the same account
        invoice_account = invoice.project_details.project.account
        invoice_year, invoice_month = invoice.start.year, invoice.start.month
        invoices_to_delete: List[Invoice] = get_account_invoices(invoice_account, invoice_year, invoice_month)
        for invoice in invoices_to_delete:
            self.do_delete_invoice(invoice, request)

    def do_delete_invoice(self, invoice: Invoice, request: HttpRequest):
        if hasattr(invoice.project_details.project, "projectprepaymentdetail"):
            invoice.project_details.project.projectprepaymentdetail.restore_funds(invoice, request)
        invoice.delete()
        messages.success(request, f"Invoice {invoice.invoice_number} was successfully deleted.", "data-speed=30000")

    def get_deleted_objects(self, invoices: Iterable[Invoice], request: HttpRequest, admin_site):
        # gather all related objects to delete: other invoices, cap discounts, cap discount amounts
        related_invoices = []
        for invoice in invoices:
            invoice_account = invoice.project_details.project.account
            invoice_year, invoice_month = invoice.start.year, invoice.start.month
            related_invoices.extend(get_account_invoices(invoice_account, invoice_year, invoice_month))
        invoice_deleted_objects = get_deleted_objects(related_invoices, request, admin_site)
        return (
            invoice_deleted_objects[0],
            {**invoice_deleted_objects[1]},
            invoice_deleted_objects[2],
            invoice_deleted_objects[2],
        )

    def _tool_usages(self, start, end, config, x_filter=Q(), raise_no_rate=True) -> List[BillableItem]:
        usage_events = UsageEvent.objects.filter(end__gte=start, end__lte=end).prefetch_related(
            "user",
            "operator",
            "waived_by",
            "tool",
            "tool__core_rel__core_facility",
            "project",
            "project__account",
            "project__projectbillingdetails__category",
            "project__projectbillingdetails__institution__institution_type",
            "project__projectbillingdetails__department",
        )
        usage_events = usage_events.filter(x_filter).order_by("start")
        return flatten([process_billables(u, u.project, config, raise_no_rate) for u in usage_events])

    def _area_access_records(self, start, end, config, x_filter=Q(), raise_no_rate=True) -> List[BillableItem]:
        access_records = AreaAccessRecord.objects.filter(end__gte=start, end__lte=end).prefetch_related(
            "project__account",
            "project__projectbillingdetails__category",
            "project__projectbillingdetails__institution__institution_type",
            "project__projectbillingdetails__department",
            "customer",
            "staff_charge__staff_member",
            "waived_by",
            "area",
            "area__core_rel__core_facility",
        )
        access_records = access_records.filter(x_filter).order_by("start")
        return flatten([process_billables(access, access.project, config, raise_no_rate) for access in access_records])

    def _missed_reservations(self, start, end, config, x_filter=Q(), raise_no_rate=True) -> List[BillableItem]:
        missed_res = Reservation.objects.filter(missed=True, end__gte=start, end__lte=end).prefetch_related(
            "user",
            "waived_by",
            "project",
            "project__account",
            "project__projectbillingdetails__category",
            "project__projectbillingdetails__institution__institution_type",
            "project__projectbillingdetails__department",
            "tool",
            "tool__core_rel__core_facility",
            "area",
            "area__core_rel__core_facility",
        )
        missed_res = missed_res.filter(x_filter).order_by("start")
        return flatten([process_billables(missed, missed.project, config, raise_no_rate) for missed in missed_res])

    def _staff_charges(self, start, end, config, x_filter=Q(), raise_no_rate=True) -> List[BillableItem]:
        staff_charges = StaffCharge.objects.filter(end__gte=start, end__lte=end).prefetch_related(
            "project",
            "project__account",
            "project__projectbillingdetails__category",
            "project__projectbillingdetails__institution__institution_type",
            "project__projectbillingdetails__department",
            "customer",
            "waived_by",
            "staff_member",
            "core_rel__core_facility",
        )
        staff_charges = staff_charges.filter(x_filter).order_by("start")
        return flatten([process_billables(c, c.project, config, raise_no_rate) for c in staff_charges])

    def _consumable_withdrawals(self, start, end, config, x_filter=Q(), raise_no_rate=True) -> List[BillableItem]:
        withdrawals = ConsumableWithdraw.objects.filter(date__gte=start, date__lte=end).prefetch_related(
            "project",
            "project__account",
            "project__projectbillingdetails__category",
            "project__projectbillingdetails__institution__institution_type",
            "project__projectbillingdetails__department",
            "customer",
            "merchant",
            "waived_by",
            "consumable",
            "consumable__core_rel__core_facility",
        )
        withdrawals = withdrawals.filter(x_filter).order_by("date")
        return flatten([process_billables(w, w.project, config, raise_no_rate) for w in withdrawals])

    def _training_sessions(self, start, end, config, x_filter=Q(), raise_no_rate=True) -> List[BillableItem]:
        training_sessions = TrainingSession.objects.filter(date__gte=start, date__lte=end).prefetch_related(
            "trainee",
            "trainer",
            "waived_by",
            "tool",
            "tool__core_rel__core_facility",
            "project",
            "project__account",
            "project__projectbillingdetails__category",
            "project__projectbillingdetails__institution__institution_type",
            "project__projectbillingdetails__department",
        )
        training_sessions = training_sessions.filter(x_filter).order_by("date")
        return flatten([process_billables(t, t.project, config, raise_no_rate) for t in training_sessions])

    def _custom_charges(self, start, end, config, x_filter=Q(), raise_no_rate=True) -> List[BillableItem]:
        custom_charges = CustomCharge.objects.filter(date__gte=start, date__lte=end).prefetch_related(
            "project",
            "project__account",
            "project__projectbillingdetails__category",
            "project__projectbillingdetails__institution__institution_type",
            "project__projectbillingdetails__department",
            "customer",
            "waived_by",
            "core_facility",
        )
        custom_charges = custom_charges.filter(x_filter).order_by("date")
        return flatten([process_billables(cc, cc.project, config, raise_no_rate) for cc in custom_charges])

    def get_invoice_summary_items(self, invoice, detail_items: List[InvoiceDetailItem]) -> List[InvoiceSummaryItem]:
        summaries: List[InvoiceSummaryItem] = []
        # Core facilities sorted alphabetically by non-empty ones first
        for core_facility in invoice.sorted_core_facilities(detail_items):
            details = [item for item in detail_items if item.core_facility == core_facility]
            summaries.extend(self.get_summary_items_for_facility(invoice, core_facility, details))

        # Recap of all charges
        charges_amount = sum(
            [
                summary.amount
                for summary in summaries
                if summary.summary_item_type == InvoiceSummaryItem.InvoiceSummaryItemType.SUBTOTAL
            ]
        )

        # Tax
        tax_amount = Decimal(0)
        if invoice.configuration.tax and charges_amount > Decimal(0) and not invoice.project_details.no_tax:
            tax = InvoiceSummaryItem(
                invoice=invoice, name=f"{invoice.configuration.tax_name} ({invoice.configuration.tax_display()}%)"
            )
            tax.summary_item_type = InvoiceSummaryItem.InvoiceSummaryItemType.TAX
            tax_amount = round_decimal_amount(charges_amount * invoice.configuration.tax_amount())
            tax.amount = tax_amount
            summaries.append(tax)

        invoice.total_amount = charges_amount + tax_amount

        if hasattr(invoice.project_details.project, "projectprepaymentdetail"):
            prepayment = invoice.project_details.project.projectprepaymentdetail
            fund_summaries = prepayment.invoice_fund_summaries(invoice)
            total_fund_used = Decimal(0)
            for fund_summary in fund_summaries:
                summaries.append(fund_summary)
                total_fund_used += fund_summary.amount
            invoice.total_amount = invoice.total_amount + total_fund_used
        if InvoiceCustomization.get_bool("invoice_show_hard_cap_status"):
            hard_caps = (
                invoice.project_details.project.projectbillinghardcap_set.filter(enabled=True)
                .exclude(start_date__lt=invoice.start, end_date__lt=invoice.start)
                .exclude(start_date__gt=invoice.end, end_date__gt=invoice.end)
            )
            for hard_cap in hard_caps:
                start_datetime = make_aware(datetime.combine(hard_cap.start_date or date.min, time.min))
                end_datetime = make_aware(datetime.combine(hard_cap.end_date or date.max, time.max))
                date_range_display = ""
                if hard_cap.start_date and hard_cap.end_date:
                    date_range_display = f" during the period {format_daterange(hard_cap.start_date, hard_cap.end_date, d_format='SHORT_DATE_FORMAT')}"
                elif hard_cap.start_date:
                    date_range_display = f" since {format_datetime(hard_cap.start_date, df='SHORT_DATE_FORMAT')}"
                elif hard_cap.end_date:
                    date_range_display = f" until {format_datetime(hard_cap.end_date, df='SHORT_DATE_FORMAT')}"
                charges, amount = get_charges_amount_between(
                    hard_cap.project,
                    hard_cap.configuration,
                    start_datetime,
                    end_datetime,
                    hard_cap.billable_charge_types,
                )
                hard_cap_summary = InvoiceSummaryItem(
                    invoice=invoice,
                    name=f"CAP - You have used: {display_amount(amount, invoice.configuration)} of {display_amount(hard_cap.amount, invoice.configuration)} available{date_range_display}",
                )
                # hard_cap_summary.amount = 0
                hard_cap_summary.summary_item_type = InvoiceSummaryItem.InvoiceSummaryItemType.OTHER
                summaries.append(hard_cap_summary)
        return summaries

    def get_summary_items_for_facility(self, invoice: Invoice, core_facility: str, details) -> List[InvoiceSummaryItem]:
        facility_summaries: List[InvoiceSummaryItem] = []
        for billable_type in BillableItemType:
            items = [item for item in details if item.item_type == billable_type.value]
            facility_summaries.extend(self.get_recap_usage_summary(invoice, core_facility, items))

        facility_subtotal = InvoiceSummaryItem(invoice=invoice, name="Subtotal", core_facility=core_facility)
        facility_subtotal.summary_item_type = InvoiceSummaryItem.InvoiceSummaryItemType.SUBTOTAL
        facility_subtotal.amount = round_decimal_amount(sum(item.amount for item in details if not item.waived))
        facility_summaries.append(facility_subtotal)

        return facility_summaries

    def get_recap_usage_summary(self, invoice, facility, items: List[InvoiceDetailItem]) -> List[InvoiceSummaryItem]:
        summary_items: List[InvoiceSummaryItem] = []
        if items:
            item_type_value = items[0].item_type
            item_names = list({item.name for item in items})
            item_names.sort()
            for item_name in item_names:
                item_with_name_list = [item for item in items if item.name == item_name]
                non_waived_items = [item for item in item_with_name_list if not item.waived]
                item_rate = item_with_name_list[0].rate
                total_q = sum(item.quantity for item in non_waived_items)
                if BillableItemType(item_type_value).is_time_type():
                    quantity_display = f" ({total_q/60:.2f} hours)"
                elif item_type_value == BillableItemType.CUSTOM_CHARGE.value:
                    quantity_display = ""
                else:
                    quantity_display = f" (x {total_q})"
                summary_item_name = f"{item_name}{quantity_display}"
                summary_item = InvoiceSummaryItem(invoice=invoice, name=summary_item_name, core_facility=facility)
                summary_item.summary_item_type = InvoiceSummaryItem.InvoiceSummaryItemType.ITEM
                summary_item.item_type = item_type_value
                summary_item.details = item_rate
                summary_item.amount = round_decimal_amount(sum(item.amount for item in non_waived_items))
                summary_items.append(summary_item)
        return summary_items

    def category_name_for_item_type(self, item_type: Optional[Union[BillableItemType, int]]) -> str:
        billable_item_type = (
            item_type
            if isinstance(item_type, BillableItemType)
            else BillableItemType(item_type) if isinstance(item_type, int) else None
        )
        if not billable_item_type:
            return ""
        if billable_item_type == BillableItemType.TOOL_USAGE:
            return "Tool Usage"
        elif billable_item_type == BillableItemType.AREA_ACCESS:
            return "Area Access"
        elif billable_item_type == BillableItemType.CONSUMABLE:
            return "Supplies/Materials"
        elif billable_item_type == BillableItemType.STAFF_CHARGE:
            return "Technical Work"
        elif billable_item_type == BillableItemType.TRAINING:
            return "Training"
        elif billable_item_type == BillableItemType.MISSED_RESERVATION:
            return "Missed Reservations"
        elif billable_item_type == BillableItemType.CUSTOM_CHARGE:
            return "Other"

    def name_for_item(self, billable_item: BillableItem) -> str:
        name = (
            getattr(billable_item.tool, "name", None)
            or getattr(billable_item.area, "name", None)
            or getattr(billable_item.consumable, "name", None)
            or getattr(billable_item.item, "name", None)
        )
        if isinstance(billable_item.item, TrainingSession):
            name = f"{billable_item.tool.name} ({billable_item.item.get_type_display()})"
        if not name and isinstance(billable_item.item, StaffCharge):
            name = "Staff time"
        return name


def billable_to_invoice_detail_item(item: BillableItem, invoice: Optional[Invoice]) -> InvoiceDetailItem:
    invoice_item = InvoiceDetailItem(invoice=invoice)
    invoice_item.content_object = item.item
    invoice_item.quantity = item.quantity
    invoice_item.start = item.start
    invoice_item.end = item.end
    invoice_item.user = item.user.username if item.user else None
    invoice_item.amount = item.amount
    invoice_item.rate = item.display_rate
    invoice_item.core_facility = item.core_facility.name if item.core_facility else None
    invoice_item.item_type = item.item_type.value
    invoice_item.name = item.name
    invoice_item.waived = item.waived
    return invoice_item


def process_billables(item, project, configuration: InvoiceConfiguration, raise_no_rate=True) -> List[BillableItem]:
    original_billable = BillableItem(item, project, configuration)
    try:
        # Only proceed if we don't have an amount already
        if not original_billable.amount:
            # If we don't, find the rates and calculate the amount
            rates = find_billable_rates(original_billable, configuration)
            billables: List[BillableItem] = []
            break_up_and_add(billables, rates, original_billable.start, original_billable.end, original_billable)
            billables = check_and_process_one_time_fees(billables, original_billable)
            return billables
    except (NoRateSetException, NoProjectCategorySetException, NoProjectDetailsSetException):
        if raise_no_rate:
            raise
    # If we already have an amount, or if there is an error & we don't raise it, return the original billable
    return [original_billable]


def check_and_process_one_time_fees(billables: List[BillableItem], original_billable) -> List[BillableItem]:
    prorated_minimum_charge, prorated_service_fee = get_pro_rated_minimum_charge_and_service_fee(
        [(billable.rate, billable.quantity) for billable in billables]
    )
    total_amount = sum(billable.amount for billable in billables)
    if prorated_minimum_charge >= total_amount:
        # The quantity might not be set on the original, so set it
        if not original_billable.quantity:
            original_billable.quantity = get_minutes_between_dates(original_billable.start, original_billable.end)
        # if the total is less than the prorated minimum charge, set the minimum charge
        original_billable.amount = prorated_minimum_charge + prorated_service_fee
        # let's set the prorated fees on the rate, so users can see it
        original_billable.rate = billables[0].rate
        original_billable.rate.minimum_charge = prorated_minimum_charge
        original_billable.rate.service_fee = prorated_service_fee
        # in this case there is no need to keep the multiple split charges, return the original
        return [original_billable]
    else:
        # otherwise we don't need the minimum charge at all, just apply the service fee to the first billable
        # and set the other ones to 0
        for i, item in enumerate(billables):
            if i == 0:
                billables[i].amount += prorated_service_fee
                billables[i].rate.service_fee = prorated_service_fee
            else:
                billables[i].rate.service_fee = Decimal(0)
        return billables


def break_up_and_add(
    billables: List[BillableItem],
    rates: List[Rate],
    start_time: datetime,
    end_time: datetime,
    original_billable: BillableItem,
):
    # This function will break the date range into chunks of BillableItems if there are multiple rates that apply
    # 1. Find the earliest rate that applies to this date range as well as its start time and end time.
    # 2. The new start is the one given by the first_rate_that_applies method (max of rate start time and original start time)
    # 3. If the new start is later than the start, that means we have a gap and need to apply the base rate to the in between times
    # 4. The new end is the min of rate start time + rate duration and the original end
    # 5. Apply it to a new BillableItem and calculate the new amount.
    # 6. Call recursively on the rest of the date range (new end until original end)
    # 7. Recursion ends when we don't have a rate with time that applies (set base rate and done)
    # use UTC to avoid issues with DST
    start_utc, end_utc = (
        start_time.astimezone(timezone.utc),
        end_time.astimezone(timezone.utc),
    )
    rate, new_start_utc, new_end_utc = earliest_rate_match(rates, start_utc, end_utc)
    base_rate = next(r for r in rates if not r.time)
    if not rate:
        billables.append(original_billable.calculate_amount(base_rate, start_utc, end_utc, copy_original=False))
    else:
        # If the new start is later than the current start, we have a "hole" between timed rates, so apply the original base rate
        if new_start_utc > start_utc:
            billables.append(original_billable.calculate_amount(base_rate, start_utc, new_start_utc))
        billables.append(original_billable.calculate_amount(rate, new_start_utc, new_end_utc))
        # keep going until we reach the end
        if new_end_utc != end_time:
            break_up_and_add(billables, rates, new_end_utc, end_time, original_billable)


def earliest_rate_match(rates: List[Rate], start_time: datetime, end_time: datetime) -> (Rate, datetime, datetime):
    new_rate, new_start, new_end = None, None, None
    for rate in rates:
        # We only care about rate with times, since they are the only ones that can override the base rate
        if rate.time:
            tmp_start, tmp_end = rate.time.earliest_match(start_time, end_time)
            if tmp_start and (not new_start or tmp_start < new_start):
                new_start = tmp_start
                new_end = tmp_end
                new_rate = rate
    return new_rate, new_start, new_end


def split_records_by_day(item: BillableItem) -> List[BillableItem]:
    split_records = []
    curr_start = localtime(item.start)

    while curr_start < item.end:
        potential_end = min(make_aware(datetime.combine(curr_start.date(), time.min) + timedelta(days=1)), item.end)
        if potential_end < item.end:
            curr_end = potential_end
        else:
            curr_end = item.end
        # Create a new record if dates are different
        if curr_end != item.end or curr_start != item.start:
            new_record = copy(item)
            new_record.start = curr_start
            new_record.end = curr_end
            new_record.quantity = get_minutes_between_dates(new_record.start, new_record.end)
            split_records.append(new_record)
        else:
            split_records.append(item)

        curr_start = curr_end
    return split_records


def get_account_invoices(account: Account, year: int, month: int) -> List[Invoice]:
    return Invoice.objects.filter(start__year=year, start__month=month, project_details__project__account=account)


def get_invoice_data_processor_class() -> InvoiceDataProcessor:
    processor_class = getattr(
        settings, "INVOICE_DATA_PROCESSOR_CLASS", "NEMO_billing.invoices.processors.InvoiceDataProcessor"
    )
    assert isinstance(processor_class, str)
    pkg, attr = processor_class.rsplit(".", 1)
    ret = getattr(importlib.import_module(pkg), attr)
    return ret()


invoice_data_processor_class = get_invoice_data_processor_class()



# Generated by Django 4.2.23 on 2025-10-03 16:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rates", "0009_rate_effective_date"),
    ]

    operations = [
        migrations.AddField(
            model_name="rate",
            name="service_fee",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="The service fee for this rate (always applies)",
                max_digits=8,
                null=True,
            ),
        ),
    ]



# Generated by Django 4.2.11 on 2024-07-26 22:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rates", "0005_version_2_3_0"),
    ]

    operations = [
        migrations.AddField(
            model_name="rate",
            name="daily_split_multi_day_charges",
            field=models.BooleanField(
                default=False,
                help_text="Check this box to split charges spanning multiple days, leave unchecked to keep one long charge",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="rate",
            name="daily_split_multi_day_charges",
            field=models.BooleanField(
                default=True,
                help_text="Check this box to split charges spanning multiple days, leave unchecked to keep one long charge",
            ),
        ),
    ]



from collections import defaultdict
from decimal import Decimal
from logging import getLogger
from typing import Dict, List

from NEMO.decorators import accounting_or_user_office_or_manager_required, any_staff_required
from NEMO.models import Account, AccountType, AdjustmentRequest, Project, ProjectType, StaffCharge, User
from NEMO.utilities import (
    BasicDisplayTable,
    export_format_datetime,
    format_datetime,
    get_day_timeframe,
    get_month_timeframe,
)
from NEMO.views.customization import AdjustmentRequestsCustomization, ProjectsAccountsCustomization
from NEMO.views.usage import date_parameters_dictionary, get_managed_projects, get_project_applications
from django.contrib.auth.decorators import login_required
from django.db.models import F, Q
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.views.decorators.http import require_GET

from NEMO_billing.customization import BillingCustomization
from NEMO_billing.invoices.models import InvoiceConfiguration, InvoiceDetailItem
from NEMO_billing.invoices.processors import BillableItem, invoice_data_processor_class as data_processor
from NEMO_billing.invoices.utilities import display_amount
from NEMO_billing.templatetags.billing_tags import cap_discount_installed

logger = getLogger(__name__)


@login_required
@require_GET
def user_usage(request):
    user: User = request.user
    user_managed_projects = get_managed_projects(user)
    customer_filter = Q(customer=user) | Q(project__in=user_managed_projects)
    user_filter = Q(user=user) | Q(project__in=user_managed_projects)
    trainee_filter = Q(trainee=user) | Q(project__in=user_managed_projects)
    csv_export = bool(request.GET.get("csv", False))
    show_only_my_usage = user_managed_projects and request.GET.get("show_only_my_usage", "enabled") == "enabled"
    if show_only_my_usage:
        # Forcing to be user only
        customer_filter &= Q(customer=user)
        user_filter &= Q(user=user)
        trainee_filter &= Q(trainee=user)
    return usage(
        request,
        usage_filter=user_filter,
        area_access_filter=customer_filter,
        staff_charges_filter=customer_filter,
        consumable_filter=customer_filter,
        reservation_filter=user_filter,
        training_filter=trainee_filter,
        custom_charges_filter=customer_filter,
        show_only_my_usage=show_only_my_usage,
        csv_export=csv_export,
        user_managed_projects=user_managed_projects,
    )


@any_staff_required
@require_GET
def staff_usage(request):
    user: User = request.user
    usage_filter = Q(operator=user) & ~Q(user=F("operator"))
    area_access_filter = Q(staff_charge__staff_member=user)
    staff_charges_filter = Q(staff_member=user)
    consumable_filter = Q(merchant=user)
    user_filter = Q(pk__in=[])
    trainee_filter = Q(trainer=user)
    custom_charges_filter = Q(creator=user)
    csv_export = bool(request.GET.get("csv", False))
    return usage(
        request,
        usage_filter=usage_filter,
        area_access_filter=area_access_filter,
        staff_charges_filter=staff_charges_filter,
        consumable_filter=consumable_filter,
        reservation_filter=user_filter,
        training_filter=trainee_filter,
        custom_charges_filter=custom_charges_filter,
        show_only_my_usage=None,
        csv_export=csv_export,
        user_managed_projects=set(),
    )


def usage(
    request,
    usage_filter,
    area_access_filter,
    staff_charges_filter,
    consumable_filter,
    reservation_filter,
    training_filter,
    custom_charges_filter,
    show_only_my_usage,
    csv_export,
    user_managed_projects,
):
    user = request.user
    base_dictionary, start, end, kind, identifier = date_parameters_dictionary(request, get_month_timeframe)
    base_dictionary["run_data_collapse"] = set_run_data_collapse(request)
    # Preloading user's managed projects'
    base_dictionary["user"] = User.objects.filter(id=user.id).prefetch_related("managed_projects").first()
    project_id = request.GET.get("project") or request.GET.get("pi_project")
    if user_managed_projects:
        base_dictionary["selected_project"] = "all"
    if project_id:
        project = get_object_or_404(Project, id=project_id)
        if request.GET.get("project"):
            base_dictionary["selected_user_project"] = project
        else:
            base_dictionary["selected_project"] = project
            base_dictionary["explicitly_display_customer"] = True
        area_access_filter &= Q(project=project)
        staff_charges_filter &= Q(project=project)
        usage_filter &= Q(project=project)
        consumable_filter &= Q(project=project)
        reservation_filter &= Q(project=project)
        training_filter &= Q(project=project)
        custom_charges_filter &= Q(project=project)
    config = InvoiceConfiguration.first_or_default()
    detailed_items = sorted_billable_items(
        start,
        end,
        config,
        usage_filter,
        area_access_filter,
        staff_charges_filter,
        consumable_filter,
        reservation_filter,
        training_filter,
        custom_charges_filter,
    )
    if csv_export:
        return csv_export_response(user, detailed_items)
    else:
        non_waived_items = [item for item in detailed_items if not item.waived]
        total_charges = Decimal(sum(item.billable_amount for item in non_waived_items))
        dictionary = {
            "detailed_items": detailed_items,
            "total_charges": display_amount(total_charges, config),
            "can_export": True,
            "adjustment_time_limit": AdjustmentRequestsCustomization.get_date_limit(),
            "existing_adjustments": get_existing_adjustments(user),
            "charges_projects": Project.objects.filter(
                id__in=set(
                    list(user.active_projects().values_list("id", flat=True))
                    + list({item.project.id for item in detailed_items if item.project})
                )
            ),
        }
        if BillingCustomization.get_bool("billing_usage_show_pending_vs_final"):
            dictionary["pending_charges"] = display_amount(
                total_charges - sum(item.invoiced_amount or 0 for item in non_waived_items), config
            )
        if user_managed_projects:
            dictionary["pi_projects"] = user_managed_projects
            dictionary["show_only_my_usage"] = show_only_my_usage
        dictionary["no_charges"] = not dictionary["detailed_items"]
        if cap_discount_installed():
            from NEMO_billing.cap_discount.models import CAPDiscount

            if CAPDiscount.objects.filter(user=user).exists():
                dictionary["cap_discounts_url"] = reverse("usage_cap_discounts")
        if user_managed_projects and any(
            [getattr(project, "projectprepaymentdetail", None) for project in user_managed_projects if project.active]
        ):
            dictionary["project_prepayments_url"] = reverse("usage_project_prepayments")

        return render(request, "invoices/usage.html", {**base_dictionary, **dictionary})


@accounting_or_user_office_or_manager_required
@require_GET
def project_usage(request):
    base_dictionary, start, end, kind, identifier = date_parameters_dictionary(request, get_day_timeframe)
    base_dictionary["run_data_collapse"] = set_run_data_collapse(request)
    # Preloading user's managed projects'
    base_dictionary["user"] = User.objects.filter(id=request.user.id).prefetch_related("managed_projects").first()

    detailed_items: List[BillableItem] = []
    config = InvoiceConfiguration.first_or_default()

    projects = []
    user = None
    account = None
    selection = ""

    # Get selection as strings.
    selected_account_type = request.GET.get("account_type")
    selected_project_type = request.GET.get("project_type")

    try:
        if kind == "projectapplication":
            projects = Project.objects.filter(application_identifier=identifier)
            selection = identifier
        elif kind == "project":
            projects = [Project.objects.get(id=identifier)]
            selection = projects[0].name
        elif kind == "account":
            account = Account.objects.get(id=identifier)
            projects = Project.objects.filter(account=account)
            selection = account.name
        elif kind == "user":
            user = User.objects.get(id=identifier)
            projects = user.active_projects()
            selection = str(user)

        customer_filter = Q()
        user_filter = Q()
        trainee_filter = Q()
        if projects:
            customer_filter = customer_filter & Q(project__in=projects)
            user_filter = user_filter & Q(project__in=projects)
            trainee_filter = trainee_filter & Q(project__in=projects)
        if user:
            customer_filter = customer_filter & Q(customer=user)
            user_filter = user_filter & Q(user=user)
            trainee_filter = trainee_filter & Q(trainee=user)
        if selected_account_type:
            # Get a subset of projects and filter the other records using that subset.
            projects_by_account_type = Project.objects.filter(account__type__id=selected_account_type)
            customer_filter = customer_filter & Q(project__in=projects_by_account_type)
            user_filter = user_filter & Q(project__in=projects_by_account_type)
            trainee_filter = trainee_filter & Q(project__in=projects_by_account_type)
        if selected_project_type:
            # Get a subset of projects and filter the other records using that subset.
            projects_by_type = Project.objects.filter(project_types__id=selected_project_type)
            customer_filter = customer_filter & Q(project__in=projects_by_type)
            user_filter = user_filter & Q(project__in=projects_by_type)
            trainee_filter = trainee_filter & Q(project__in=projects_by_type)
        detailed_items = sorted_billable_items(
            start,
            end,
            config,
            usage_filter=user_filter,
            area_access_filter=customer_filter,
            staff_charges_filter=customer_filter,
            consumable_filter=customer_filter,
            reservation_filter=user_filter,
            training_filter=trainee_filter,
            custom_charges_filter=customer_filter,
        )
        if bool(request.GET.get("csv", False)):
            return csv_export_response(request.user, detailed_items)
    except Exception as e:
        logger.exception(e)

    # Get a list of unique account types for the dropdown field.
    account_types = AccountType.objects.filter(id__in=Account.objects.values_list("type__id", flat=True))

    # Get a list of unique project types for the dropdown field.
    project_types = ProjectType.objects.filter(id__in=Project.objects.values_list("project_types__id", flat=True))

    non_waived_items = [item for item in detailed_items if not item.waived]
    total_charges = Decimal(sum(item.billable_amount for item in non_waived_items))
    dictionary = {
        "search_items": set(Account.objects.all())
        | set(Project.objects.all())
        | set(get_project_applications())
        | set(User.objects.filter(is_active=True)),
        "detailed_items": detailed_items,
        "total_charges": display_amount(total_charges, config),
        "pending_charges": display_amount(
            total_charges - sum(item.invoiced_amount or 0 for item in non_waived_items), config
        ),
        "project_autocomplete": True,
        "adjustment_time_limit": AdjustmentRequestsCustomization.get_date_limit(),
        "existing_adjustments": get_existing_adjustments(request.user),
        "selection": selection,
        "can_export": True,
        "account_types": account_types,
        "selected_account_type": selected_account_type,
        "project_types": project_types,
        "selected_project_type": selected_project_type,
    }
    dictionary["no_charges"] = not dictionary["detailed_items"]
    if cap_discount_installed():
        from NEMO_billing.cap_discount.models import CAPDiscount

        if user and CAPDiscount.objects.filter(user=user).exists():
            dictionary["cap_discounts_url"] = reverse("usage_cap_discounts_user", args=[identifier])
        if account and CAPDiscount.objects.filter(account=account).exists():
            dictionary["cap_discounts_url"] = reverse("usage_cap_discounts_account", args=[identifier])
    return render(request, "invoices/usage.html", {**base_dictionary, **dictionary})


def sorted_billable_items(
    start,
    end,
    config,
    usage_filter,
    area_access_filter,
    staff_charges_filter,
    consumable_filter,
    reservation_filter,
    training_filter,
    custom_charges_filter,
) -> List[BillableItem]:
    items = data_processor.get_billable_items_with_charge_filters(
        start,
        end,
        config,
        usage_filter,
        area_access_filter,
        staff_charges_filter,
        consumable_filter,
        reservation_filter,
        training_filter,
        custom_charges_filter,
        False,
    )
    augment_with_invoice_items(items)
    items.sort(key=lambda x: (-x.item_type.value, x.start), reverse=True)
    return items


def augment_with_invoice_items(billables: List[BillableItem]):
    pending_vs_final = BillingCustomization.get_bool("billing_usage_show_pending_vs_final")
    # let's get invoice items in one query. We are only using item.id which means
    # we might get a few more than we really need (different types), but it's better than a query for each
    invoice_items: List[InvoiceDetailItem] = list(
        InvoiceDetailItem.objects.filter(
            invoice__voided_date__isnull=True, object_id__in=[billable.item.id for billable in billables]
        )
        .select_related("invoice__configuration", "content_type")
        .only("invoice", "content_type", "object_id", "rate", "amount", "discount", "waived")
    )
    for billable in billables:
        matching_items = [
            invoice_item
            for invoice_item in invoice_items
            if invoice_item.object_id == billable.item.id
            and invoice_item.content_type.model == billable.item._meta.model_name
        ]
        invoice_item: InvoiceDetailItem = matching_items[0] if matching_items else None
        billable.invoiced_display_amount = invoice_item.amount_display() if invoice_item else None
        billable.invoiced_amount = invoice_item.amount if invoice_item else None
        billable.invoiced_discount = invoice_item.discount if invoice_item else None
        billable.invoiced_rate = invoice_item.rate if invoice_item else None
        billable.amount = billable.amount if not billable.invoiced_amount else None
        billable.billable_rate = billable.invoiced_rate or billable.display_rate
        billable.billable_amount = billable.invoiced_amount or billable.amount or 0
        billable.billable_display_amount = billable.invoiced_display_amount or (
            f"{'(pending) ' if pending_vs_final else ''}{billable.display_amount}" if billable.display_amount else ""
        )
        billable.merged_amount = billable.invoiced_amount or billable.amount

        if hasattr(billable, "rate") and billable.rate:
            if billable.rate.daily:
                billable.unit_type = "daily"
            elif billable.rate.flat:
                billable.unit_type = "flat"
            else:
                billable.unit_type = "hourly"
        else:
            billable.unit_type = None

        if billable.unit_type == "hourly" and billable.quantity is not None:
            # Convert quantity from minutes to hours.
            billable.unit_quantity = Decimal(billable.quantity / 60).quantize(Decimal("0.01"))
        elif billable.unit_type == "flat" and billable.quantity is not None:
            # Set quantity to 1.
            billable.unit_quantity = 1
        elif billable.unit_type == "daily" and billable.quantity is not None:
            # First usage in the day gets a billable.amount, but subsequent check-outs in the same day will have `billable.amount` = 0.
            # Check-outs spanning multiple days are given separate `billable` datasets, one per day with the appropriate charge amount,
            # if `billable.rate.daily_split_multi_day_charges` is true.
            billable.unit_quantity = 1 if billable.billable_amount > 0 else 0
        else:
            # Otherwise, keep the quantity value (could be None)
            # Getting here means the `billable.unit_type` is not accounted for.
            billable.unit_quantity = billable.quantity


def csv_export_response(user: User, detailed_items: List[BillableItem]):
    table_result = BasicDisplayTable()
    table_result.add_header(("type", "Type"))
    table_result.add_header(("user", "User"))
    table_result.add_header(("username", "Username"))
    table_result.add_header(("name", "Item"))
    table_result.add_header(("project", "Project"))
    if user.is_any_part_of_staff:
        table_result.add_header(
            ("application", ProjectsAccountsCustomization.get("project_application_identifier_name"))
        )
    table_result.add_header(("start", "Start time"))
    table_result.add_header(("end", "End time"))
    table_result.add_header(("quantity", "Quantity"))
    table_result.add_header(("rate", "Rate"))
    table_result.add_header(("cost", "Cost"))
    table_result.add_header(("waived", "Waived"))
    pending_vs_final = BillingCustomization.get_bool("billing_usage_show_pending_vs_final")
    if pending_vs_final:
        table_result.add_header(("cost_pending", "Cost (Pending)"))
    if user.is_any_part_of_staff:
        table_result.add_header(("staff_charge_note", "Staff charge note"))
    for billable_item in detailed_items:
        billable_dict = {
            "type": billable_item.item_type.display_name(),
            "user": billable_item.user.get_name(),
            "username": billable_item.user.username,
            "name": billable_item.name,
            "project": billable_item.project.name if billable_item.project else "",
            "start": format_datetime(billable_item.start, "SHORT_DATETIME_FORMAT"),
            "end": format_datetime(billable_item.end, "SHORT_DATETIME_FORMAT"),
            "quantity": billable_item.quantity,
            "rate": billable_item.invoiced_rate or (billable_item.rate.display_rate() if billable_item.rate else ""),
            "cost": round(billable_item.invoiced_amount, 2) if billable_item.invoiced_amount else "",
            "waived": billable_item.waived,
            "application": billable_item.project.application_identifier if billable_item.project else "",
            "staff_charge_note": (
                billable_item.item.note if billable_item.item and isinstance(billable_item.item, StaffCharge) else ""
            ),
        }
        if pending_vs_final:
            billable_dict["cost_pending"] = round(billable_item.amount, 2) if billable_item.amount else ""
        else:
            billable_dict["cost"] = round(billable_item.merged_amount, 2) if billable_item.merged_amount else ""
        table_result.add_row(billable_dict)
    response = table_result.to_csv()
    filename = f"usage_export_{export_format_datetime()}.csv"
    response["Content-Disposition"] = f'attachment; filename="{filename}"'
    return response


def get_existing_adjustments(user) -> Dict[int, List]:
    existing_adjustments = defaultdict(list)
    for values in (
        AdjustmentRequest.objects.filter(deleted=False, creator=user).values("item_type", "item_id").distinct()
    ):
        existing_adjustments[values["item_type"]].append(values["item_id"])
    return existing_adjustments


# Remove when NEMO 7.2.0 is released
def set_run_data_collapse(request):
    if request.GET.get("run_data_collapse"):
        request.session["run_data_collapse"] = request.GET.get("run_data_collapse") == "true"
    run_data_collapse = False
    if "run_data_collapse" in request.session:
        run_data_collapse = request.session["run_data_collapse"]
    return run_data_collapse



from _decimal import Decimal
from datetime import date, datetime
from typing import List

from NEMO.models import (
    Account,
    Project,
)
from NEMO.utilities import RecurrenceFrequency, new_model_copy
from dateutil.relativedelta import relativedelta
from django.core.exceptions import ValidationError
from django.db import IntegrityError, transaction
from django.forms import CheckboxInput, HiddenInput
from django.test import TestCase
from django.utils import timezone

from NEMO_billing.admin import CustomChargeAdminForm
from NEMO_billing.cap_discount.models import CAPDiscount, CAPDiscountAmount, CAPDiscountConfiguration, CAPDiscountTier
from NEMO_billing.invoices.models import BillableItemType, Invoice, InvoiceConfiguration
from NEMO_billing.invoices.processors import invoice_data_processor_class as processor
from NEMO_billing.models import CoreFacility, CustomCharge
from NEMO_billing.rates.models import RateCategory
from NEMO_billing.tests.cap.test_cap_utilities import (
    add_charges_of_type,
    create_items,
    delete_all_billables,
    get_other_billable_types,
)
from NEMO_billing.tests.test_utilities import basic_data


class TestCapAmount(TestCase):
    def setUp(self) -> None:
        self.user, self.project, self.tool, self.area = basic_data(set_rate_category=True)
        cap_configuration = CAPDiscountConfiguration(rate_category=self.project.projectbillingdetails.category)
        cap_configuration.reset_frequency = RecurrenceFrequency.YEARLY.value
        cap_configuration.start = date.today()
        cap_configuration.start_on_first_charge = False
        self.assertRaises(ValidationError, cap_configuration.full_clean)
        cap_configuration.charge_types = BillableItemType.TOOL_USAGE.value
        cap_configuration.full_clean()
        cap_configuration.save()
        cap_configuration.capdiscounttier_set.set(
            [
                CAPDiscountTier.objects.create(
                    cap_discount_configuration=cap_configuration, amount=Decimal("100"), discount=Decimal("100")
                )
            ]
        )
        self.cap_configuration: CAPDiscountConfiguration = CAPDiscountConfiguration.objects.get(pk=cap_configuration.pk)
        self.tool, self.area, self.consumable, self.staff = create_items()

    def test_cap_config_required_fields(self):
        cap_config = self.cap_configuration
        # right now it's clean
        cap_config.full_clean()
        # No limit amount
        cap_tier = cap_config.capdiscounttier_set.first()
        cap_tier.amount = None
        self.assertRaises(ValidationError, cap_tier.full_clean)
        # limit of 0
        cap_tier.amount = Decimal(0)
        self.assertRaises(ValidationError, cap_tier.full_clean)
        cap_tier.amount = Decimal("1")
        cap_tier.full_clean()
        # No discount
        cap_tier.discount = None
        self.assertRaises(ValidationError, cap_tier.full_clean)
        # This is intentional, since it could be a higher tier with no discount
        cap_tier.discount = Decimal(0)
        cap_tier.full_clean()
        # Cannot have more than 100% discount
        cap_tier.discount = Decimal("101")
        self.assertRaises(ValidationError, cap_tier.full_clean)
        # 100% ok
        cap_tier.discount = Decimal("100")
        cap_tier.full_clean()
        # Start date is required when not starting on first eligible charge
        cap_config.start = None
        cap_config.start_on_first_charge = False
        self.assertRaises(ValidationError, cap_config.full_clean)
        cap_config.start = datetime.now()
        cap_config.full_clean()
        # Year set but not month
        cap_config.start_year = 2022
        cap_config.start_month = None
        self.assertRaises(ValidationError, cap_config.full_clean)
        # Month set but not year
        cap_config.start_year = None
        cap_config.start_month = 2
        self.assertRaises(ValidationError, cap_config.full_clean)
        # Start not set but start of first charge is -> OK
        cap_config.start = None
        cap_config.start_on_first_charge = True
        cap_config.full_clean()
        # Now get a CAP and check required fields
        cap = cap_config.get_or_create_cap_discount(
            account=self.project.account, username=self.user.username, start=date.today()
        )
        # override in the past not okay
        cap.next_reset_override = date.today() - relativedelta(months=1)
        self.assertRaises(ValidationError, cap.full_clean)
        # today should be ok
        cap.next_reset_override = date.today()
        cap.full_clean()
        # future is ok
        cap.next_reset_override = date.today() + relativedelta(months=14)
        cap.full_clean()
        # No account
        cap.account = None
        self.assertRaises(ValidationError, cap.full_clean)
        cap.account = self.project.account
        cap.full_clean()
        # User has to belong to account
        new_account = Account.objects.create(name="new account")
        cap.account = new_account
        self.assertRaises(ValidationError, cap.full_clean)

    def test_cap_uniqueness(self):
        # Simulate having a cap
        cap = self.cap_configuration.get_or_create_cap_discount(
            account=self.project.account, username=self.user.username, start=date.today()
        )
        # This new cap has same user/account combination as cap (with no core facility)
        new_cap = CAPDiscount(
            configuration=self.cap_configuration,
            user=self.user,
            account=self.project.account,
            charge_types=BillableItemType.TOOL_USAGE.value,
            reset_frequency=RecurrenceFrequency.YEARLY.value,
        )
        with transaction.atomic():
            self.assertRaises(IntegrityError, new_cap.save)
        new_project = Project.objects.create(name="p", account=Account.objects.create(name="new account"))
        self.user.projects.add(new_project)
        new_cap.account = new_project.account
        new_cap.save()
        CAPDiscountTier.objects.create(cap_discount=new_cap, amount=Decimal("5"), discount=Decimal("50"))
        facility = CoreFacility.objects.create(name="facility_1")
        # Set facility on original
        self.cap_configuration.core_facilities.set([facility])
        # Create new config without facility
        new_config = new_model_copy(self.cap_configuration)
        new_config.save()
        # Save, no problems
        new_config.core_facilities.clear()
        # Copy cap but without facility, should be fine
        new_cap.configuration = new_config
        new_cap.account = self.project.account
        new_cap.user = self.user
        new_cap.save()
        # Set same facility, should break
        with transaction.atomic():
            self.assertRaises(IntegrityError, new_config.core_facilities.set, [facility])
        # Set one of the same facilities, should break
        facility_2 = CoreFacility.objects.create(name="facility_2")
        with transaction.atomic():
            self.assertRaises(IntegrityError, new_config.core_facilities.set, [facility, facility_2])
        # Set only facility 2, should be fine
        new_config.core_facilities.set([facility_2])
        # Now try to add same facility 2 to first config, should break
        with transaction.atomic():
            self.assertRaises(IntegrityError, self.cap_configuration.core_facilities.add, facility_2)
        # Clear the first one, all good
        self.cap_configuration.core_facilities.clear()
        # Clear the second one, should break
        with transaction.atomic():
            self.assertRaises(IntegrityError, new_config.core_facilities.clear)
        # Create a new one without facilities, should break
        with transaction.atomic():
            self.assertRaises(
                IntegrityError,
                CAPDiscountConfiguration.objects.create,
            )

    def test_reset(self):
        # Simulate previously created CAP
        cap = self.cap_configuration.get_or_create_cap_discount(
            account=self.project.account, username=self.user.username, start=date.today()
        )
        # Change a few things in the cap configuration, the cap itself should not change until reset
        cap_configuration = self.cap_configuration
        cap_tier = cap_configuration.capdiscounttier_set.first()
        cap_tier.amount = Decimal(200)
        cap_tier.discount = Decimal(50)
        cap_tier.save()
        cap_configuration.charge_types = BillableItemType.AREA_ACCESS.value
        cap_configuration.reset_interval = 2
        cap_configuration.reset_frequency = RecurrenceFrequency.MONTHLY.value
        cap_configuration.save()
        # Check before reset
        cap = self.cap_configuration.get_or_create_cap_discount(
            account=self.project.account, username=self.user.username, start=date.today()
        )
        self.assertEqual(cap.capdiscounttier_set.first().amount, 100)
        self.assertEqual(cap.capdiscounttier_set.first().discount, 100)
        self.assertEqual(cap.billable_charge_types, [BillableItemType.TOOL_USAGE])
        self.assertEqual(cap.reset_interval, 1)
        self.assertEqual(cap.reset_frequency, RecurrenceFrequency.YEARLY.value)
        # Prepare for reset with override
        cap.next_reset_override = timezone.now().date()
        cap.reset()
        # Check after reset
        cap_after_reset = self.cap_configuration.get_or_create_cap_discount(
            account=self.project.account, username=self.user.username, start=date.today()
        )
        self.assertFalse(cap_after_reset.next_reset_override)
        self.assertEqual(cap_after_reset.capdiscounttier_set.first().amount, 200)
        self.assertEqual(cap_after_reset.capdiscounttier_set.first().discount, 50)
        self.assertEqual(cap_after_reset.billable_charge_types, [BillableItemType.AREA_ACCESS])
        self.assertEqual(cap_after_reset.reset_interval, 2)
        self.assertEqual(cap_after_reset.reset_frequency, RecurrenceFrequency.MONTHLY.value)

    def test_next_reset_override(self):
        cap = self.cap_configuration.get_or_create_cap_discount(
            account=self.project.account, username=self.user.username, start=date.today()
        )
        self.assertEqual(cap.next_reset().date(), cap.start + relativedelta(years=1))
        cap.next_reset_override = cap.start + relativedelta(months=1)
        cap.save()
        cap = CAPDiscount.objects.get(pk=cap.pk)
        self.assertTrue(cap.next_reset_override)
        self.assertEqual(cap.next_reset().date(), cap.next_reset_override)
        cap.reset()
        cap = CAPDiscount.objects.get(pk=cap.pk)
        self.assertFalse(cap.next_reset_override)

    def test_next_reset_override_first_usage(self):
        self.cap_configuration.start_on_first_charge = True
        self.cap_configuration.save()
        cap = self.cap_configuration.get_or_create_cap_discount(
            account=self.project.account,
            username=self.user.username,
            start=date.today(),
        )
        self.assertEqual(cap.next_reset().date(), cap.start + relativedelta(years=1))
        cap.next_reset_override = cap.start + relativedelta(months=1)
        previous_override = cap.next_reset_override
        cap.save()
        cap = CAPDiscount.objects.get(pk=cap.pk)
        self.assertTrue(cap.next_reset_override)
        self.assertEqual(cap.next_reset().date(), cap.next_reset_override)
        cap.reset()
        cap = CAPDiscount.objects.get(pk=cap.pk)
        # New start date should be set to last reset override (since start on first usage is set)
        self.assertEqual(cap.start, previous_override)
        self.assertFalse(cap.next_reset_override)

    def test_custom_charge_cap_eligible_field(self):
        with self.modify_settings(INSTALLED_APPS={"prepend": "NEMO_billing.cap_discount"}):
            form = CustomChargeAdminForm()
            self.assertFalse(form.fields["cap_eligible"].disabled)
            self.assertIsInstance(form.fields["cap_eligible"].widget, CheckboxInput)
        with self.modify_settings(INSTALLED_APPS={"remove": "NEMO_billing.cap_discount"}):
            form = CustomChargeAdminForm()
            self.assertTrue(form.fields["cap_eligible"].disabled)
            self.assertIsInstance(form.fields["cap_eligible"].widget, HiddenInput)

    def test_cap_eligible_custom_charge(self):
        facility = CoreFacility.objects.create(name="facility")
        # Add custom charge types to CAP entity
        self.cap_configuration.charge_types = [BillableItemType.TOOL_USAGE.value, BillableItemType.CUSTOM_CHARGE.value]
        self.cap_configuration.save()
        # Create cap eligible charge for different rate category. There is no matching CAPDiscount. It should break if cap discount is installed
        charge_date = timezone.now()
        self.project.projectbillingdetails.category = RateCategory.objects.get(name="Industry")
        self.project.projectbillingdetails.save()
        custom_charge = CustomCharge(
            name="custom",
            customer=self.user,
            creator=self.staff,
            project=self.project,
            date=charge_date,
            amount=10,
            cap_eligible=True,
        )
        with self.modify_settings(INSTALLED_APPS={"prepend": "NEMO_billing.cap_discount"}):
            self.assertRaises(ValidationError, custom_charge.full_clean)
            # put Academia back
            self.project.projectbillingdetails.category = RateCategory.objects.get(name="Academia")
            self.project.projectbillingdetails.save()
            custom_charge.full_clean()
            self.cap_configuration.core_facilities.set([facility])
            # No existing cap with no facility
            self.assertRaises(ValidationError, custom_charge.full_clean)
            custom_charge.core_facility = facility
            # now ok
            custom_charge.full_clean()
            # reset
            self.cap_configuration.core_facilities.clear()
        with self.modify_settings(INSTALLED_APPS={"remove": "NEMO_billing.cap_discount"}):
            # reset
            custom_charge.core_facility = None
            # Even with staff as custom charge, it doesn't matter since cap_discount is not installed
            self.staff.projects.add(self.project)
            custom_charge.customer = self.staff
            custom_charge.full_clean()
            # also fine with different facility
            custom_charge.core_facility = facility
            custom_charge.full_clean()

    def test_cap_amounts(self):
        # test uniqueness
        # get or create will initialize an amount for us
        cap = self.cap_configuration.get_or_create_cap_discount(
            account=self.project.account, username=self.user.username, start=date.today()
        )
        # Should break because it's not unique (month/year)
        new_amount = CAPDiscountAmount(
            cap_discount=cap, year=date.today().year, month=date.today().month, start=0, end=100
        )
        self.assertRaises(ValidationError, new_amount.full_clean)
        # All good with date of last month
        last_month = date.today() - relativedelta(months=1)
        new_amount = CAPDiscountAmount(cap_discount=cap, year=last_month.year, month=last_month.month, start=0, end=100)
        new_amount.full_clean()
        # Copy previous cap and re-save it, should fail this time with unique issue on CAP user/account combination
        new_cap = new_model_copy(cap)
        with transaction.atomic():
            self.assertRaises(IntegrityError, new_cap.save)
        # All good now with same date but different cap discount object
        new_cap.account = Account.objects.create(name="new acc")
        new_cap.save()
        new_amount = CAPDiscountAmount(
            cap_discount=new_cap, year=last_month.year, month=last_month.month, start=0, end=100
        )
        new_amount.full_clean()

    def test_set_start_not_needed(self):
        # If start date, nothing happens
        cap_configuration = self.cap_configuration
        cap_configuration.start = date.today()
        cap_configuration.save()
        add_charges_of_type(self, [BillableItemType.TOOL_USAGE], timezone.now())
        self.generate_invoice(timezone.now())
        cap_configuration = CAPDiscount.objects.get(pk=cap_configuration.pk)
        # But the date is set to today's month and year
        self.assertEqual(date.today().month, cap_configuration.start_month)
        self.assertEqual(date.today().year, cap_configuration.start_year)

    def test_all_types_of_charges(self):
        # Test each charge
        for billable_type in BillableItemType:
            with self.subTest(billable_type.name):
                self.charges_subtest([billable_type])
                delete_all_billables()
        # Test combination of charges
        for i in range(1, 8):
            for j in range(i + 1, 8):
                with self.subTest(f"{BillableItemType(i).name}-{BillableItemType(j).name}"):
                    self.charges_subtest([BillableItemType(i), BillableItemType(j)])
                    delete_all_billables()

    def charges_subtest(self, billable_charge_types: List[BillableItemType]):
        # Reset start date and amounts
        cap_configuration = self.cap_configuration
        cap_configuration.start = None
        cap_configuration.start_on_first_charge = True
        cap_configuration.charge_types = ",".join([str(t.value) for t in billable_charge_types])
        cap_configuration.full_clean()
        cap_configuration.save()
        CAPDiscountAmount.objects.filter(cap_discount__configuration=cap_configuration).delete()
        CAPDiscount.objects.filter(configuration=cap_configuration).delete()
        # If the charge end date is earlier than invoice start date, nothing happens
        Invoice.objects.all().delete()
        end = timezone.now() - relativedelta(months=1)
        self.assertTrue(end < timezone.now())
        add_charges_of_type(self, cap_configuration.billable_charge_types, end)
        self.generate_invoice(timezone.now())
        cap_qs = CAPDiscount.objects.filter(configuration=cap_configuration)
        self.assertFalse(cap_qs.exists())
        # If the charge end date is newer than invoice end date, nothing happens
        Invoice.objects.all().delete()
        two_months_ago_date = timezone.now() - relativedelta(months=2)
        self.assertTrue(end > two_months_ago_date)
        add_charges_of_type(self, cap_configuration.billable_charge_types, end)
        self.generate_invoice(two_months_ago_date)
        cap_qs = CAPDiscount.objects.filter(configuration=cap_configuration)
        self.assertFalse(cap_qs.exists())
        # If charge is a from different type, nothing happens
        # Add all other types of charges
        Invoice.objects.all().delete()
        today = timezone.now()
        add_charges_of_type(self, get_other_billable_types(cap_configuration.billable_charge_types), today)
        self.generate_invoice(today)
        cap_qs = CAPDiscount.objects.filter(configuration=cap_configuration)
        self.assertFalse(cap_qs.exists())
        # If it's an eligible charge, start date should be set to today's month
        Invoice.objects.all().delete()
        end = today
        add_charges_of_type(self, cap_configuration.billable_charge_types, end)
        self.generate_invoice(today)
        cap = CAPDiscount.objects.get(configuration=cap_configuration)
        self.assertEqual(cap.start_month, today.month)
        self.assertEqual(cap.start_year, today.year)

    def generate_invoice(self, invoice_date):
        invoice_configuration = InvoiceConfiguration.first_or_default()
        if not invoice_configuration.id:
            invoice_configuration.save()
        processor.generate_invoice_for_account(
            invoice_date.strftime("%B %Y"), self.project.account, invoice_configuration, self.user
        )

    def test_voiding_invoices(self):
        # we should only let the latest invoice be voided. then the previous one can be too, etc.
        # make sure the relevant cap amounts are voided as well or deleted?
        # the problem is that some charges already hit cap for project 2 based on project 1's charges, so it's bad to go back
        pass

    def test_project_no_cap(self):
        # Test that a charge with no_cap=True is not included in cap calculations
        charge_date = timezone.now()

        # Set no_cap=True and create a charge
        self.project.projectbillingdetails.no_cap = True
        self.project.projectbillingdetails.save()
        add_charges_of_type(self, self.cap_configuration.billable_charge_types, charge_date)
        self.generate_invoice(charge_date)

        # Check no CAP was created since charges were excluded
        cap_qs = CAPDiscount.objects.filter(configuration=self.cap_configuration)
        self.assertFalse(cap_qs.exists())

        Invoice.objects.all().delete()

        # Set no_cap=False and create another charge
        self.project.projectbillingdetails.no_cap = False
        self.project.projectbillingdetails.save()
        add_charges_of_type(self, self.cap_configuration.billable_charge_types, charge_date)
        self.generate_invoice(charge_date)

        Invoice.objects.all().delete()

        # Check CAP was created since charges were included
        cap = CAPDiscount.objects.get(configuration=self.cap_configuration)
        self.assertEqual(cap.start_month, charge_date.month)
        self.assertEqual(cap.start_year, charge_date.year)



from django.test import TestCase
from django.urls import reverse

from NEMO_billing.models import CoreFacility
from NEMO_billing.tests.test_utilities import basic_data


class TestURLs(TestCase):
    def test_staff_charge(self):
        user, project, tool, area = basic_data()
        CoreFacility.objects.create(name="Test")
        self.client.force_login(user)
        response = self.client.get(reverse("staff_charges"), follow=True)
        self.assertEqual(response.status_code, 200)
        # it's redirect to login
        self.assertEqual(response.request["PATH_INFO"], "/")
        # Make user staff
        user.is_staff = True
        user.save()
        response = self.client.get(reverse("staff_charges"), follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertNotEqual(response.request["PATH_INFO"], "/")
        self.assertContains(response, "core facility")

    def test_api(self):
        user, project, tool, area = basic_data()
        self.client.force_login(user)
        response = self.client.get(reverse("billingdata-list"), follow=True)
        # Forbidden
        self.assertEqual(response.status_code, 403)
        # Make user admin
        user.is_superuser = True
        user.save()
        response = self.client.get(reverse("billingdata-list") + "?start=01/01/2020&end=01/01/2099", follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertNotEqual(response.request["PATH_INFO"], "/")



from datetime import date

from NEMO.exceptions import ProjectChargeException
from NEMO.utilities import format_datetime


class ProjectFundsNotSetException(ProjectChargeException):
    def __init__(self, project, msg=None):
        new_msg = f"The project {project.name} does not have any funds set up."
        super().__init__(project, None, msg or new_msg)


class ProjectFundsFacilityNotAllowedException(ProjectChargeException):
    def __init__(self, project, only_facilities, msg=None):
        new_msg = f"The project {project.name} can only be used for the following core facilities: {', '.join(facility.name for facility in only_facilities)}."
        super().__init__(project, None, msg or new_msg)


class ProjectFundsExpiredException(ProjectChargeException):
    def __init__(self, project, msg=None):
        new_msg = f"All of the funds for {project.name} have expired."
        super().__init__(project, None, msg or new_msg)


class ProjectFundsInactiveException(ProjectChargeException):
    def __init__(self, project, check_date: date, msg=None):
        new_msg = f"Project {project.name} does not have any active funds for {format_datetime(check_date, 'F Y')}."
        super().__init__(project, None, msg or new_msg)


class ProjectInsufficientFundsException(ProjectChargeException):
    def __init__(self, project, msg=None):
        new_msg = f"Insufficient funds for {project.name}."
        super().__init__(project, None, msg or new_msg)



from NEMO.serializers import ModelSerializer
from NEMO.views.api import ModelViewSet, boolean_filters, datetime_filters, key_filters, number_filters, string_filters
from rest_flex_fields import FlexFieldsModelSerializer
from rest_framework.fields import CharField

from NEMO_billing.models import CoreFacility, CustomCharge, Department, Institution, InstitutionType


class CoreFacilitySerializer(ModelSerializer):
    class Meta:
        model = CoreFacility
        fields = "__all__"


class CoreFacilityViewSet(ModelViewSet):
    filename = "core_facilities"
    queryset = CoreFacility.objects.all()
    serializer_class = CoreFacilitySerializer
    filterset_fields = {
        "name": string_filters,
        "external_id": string_filters,
    }


class CustomChargeSerializer(FlexFieldsModelSerializer, ModelSerializer):
    class Meta:
        model = CustomCharge
        fields = "__all__"
        expandable_fields = {
            "customer": "NEMO.serializers.UserSerializer",
            "creator": "NEMO.serializers.UserSerializer",
            "project": "NEMO.serializers.ProjectSerializer",
            "validated_by": "NEMO.serializers.UserSerializer",
            "waived_by": "NEMO.serializers.UserSerializer",
        }

    customer_name = CharField(source="customer.get_name", read_only=True)
    creator_name = CharField(source="creator.get_name", read_only=True)


class CustomChargeViewSet(ModelViewSet):
    filename = "custom_charges"
    queryset = CustomCharge.objects.all()
    serializer_class = CustomChargeSerializer
    filterset_fields = {
        "name": string_filters,
        "customer": key_filters,
        "creator": key_filters,
        "project": key_filters,
        "date": datetime_filters,
        "amount": number_filters,
        "core_facility": key_filters,
        "validated": boolean_filters,
        "validated_by": key_filters,
        "waived": boolean_filters,
        "waived_on": datetime_filters,
        "waived_by": key_filters,
    }


class DepartmentSerializer(ModelSerializer):
    class Meta:
        model = Department
        fields = "__all__"


class DepartmentViewSet(ModelViewSet):
    filename = "departments"
    queryset = Department.objects.all()
    serializer_class = DepartmentSerializer
    filterset_fields = {
        "name": string_filters,
    }


class InstitutionTypeSerializer(ModelSerializer):
    class Meta:
        model = InstitutionType
        fields = "__all__"


class InstitutionTypeViewSet(ModelViewSet):
    filename = "institution_types"
    queryset = InstitutionType.objects.all()
    serializer_class = InstitutionTypeSerializer
    filterset_fields = {
        "name": string_filters,
    }


class InstitutionSerializer(FlexFieldsModelSerializer, ModelSerializer):
    class Meta:
        model = Institution
        fields = "__all__"
        expandable_fields = {
            "institution_type": "NEMO_billing.api.InstitutionTypeSerializer",
        }


class InstitutionViewSet(ModelViewSet):
    filename = "institutions"
    queryset = Institution.objects.all()
    serializer_class = InstitutionSerializer
    filterset_fields = {
        "name": string_filters,
        "institution_type_id": key_filters,
        "state": string_filters,
        "country": string_filters,
        "zip_code": string_filters,
    }



# Generated by Django 4.2.15 on 2024-10-05 13:39

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("NEMO_billing", "0007_alter_projectbillinghardcap_end_date_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="customcharge",
            name="validated",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="customcharge",
            name="validated_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="custom_charge_validated_set",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="customcharge",
            name="waived",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="customcharge",
            name="waived_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="custom_charge_waived_set",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="customcharge",
            name="waived_on",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="corefacility",
            name="name",
            field=models.CharField(help_text="The name of this core facility.", max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name="department",
            name="name",
            field=models.CharField(help_text="The unique name for this item", max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name="institutiontype",
            name="name",
            field=models.CharField(help_text="The unique name for this item", max_length=255, unique=True),
        ),
    ]



import datetime

from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone
from django.utils.timezone import make_aware

from NEMO_billing.rates.models import DailySchedule, RateTime


class RateTimeTestCase(TestCase):
    def test_duration(self):
        split = DailySchedule(start_time=datetime.time(20, 0, 0), end_time=datetime.time(4, 0, 0), weekday=0)
        non_split = DailySchedule(start_time=datetime.time(16, 0, 0), end_time=datetime.time(20, 0, 0), weekday=1)
        full_day = DailySchedule(weekday=1)
        full_day_time = DailySchedule(start_time=datetime.time(6, 0, 0), end_time=datetime.time(6, 0, 0), weekday=1)
        no_start_time = DailySchedule(end_time=datetime.time(6, 0, 0), weekday=1)
        no_end_time = DailySchedule(start_time=datetime.time(6, 0, 0), weekday=1)
        self.assertEqual(split.duration().total_seconds(), 8 * 3600)
        self.assertEqual(non_split.duration().total_seconds(), 4 * 3600)
        self.assertEqual(full_day.duration().total_seconds(), 24 * 3600)
        self.assertEqual(full_day_time.duration().total_seconds(), 24 * 3600)
        self.assertEqual(no_start_time.duration().total_seconds(), 6 * 3600)
        self.assertEqual(no_end_time.duration().total_seconds(), 18 * 3600)

    def test_split_time(self):
        # 4PM - 4PM => split
        d_1 = DailySchedule(weekday=0, start_time=datetime.time(16, 0, 0), end_time=datetime.time(16, 0, 0))
        self.assertTrue(d_1.is_time_range_split())
        # 4PM - 4AM => split
        d_2 = DailySchedule(weekday=0, start_time=datetime.time(16, 0, 0), end_time=datetime.time(4, 0, 0))
        self.assertTrue(d_2.is_time_range_split())
        # 12AM - 12AM => not split
        d_3 = DailySchedule(weekday=0)
        self.assertFalse(d_3.is_time_range_split())
        # 12AM - 4PM => not split
        d_4 = DailySchedule(weekday=0, end_time=datetime.time(16, 0, 0))
        self.assertFalse(d_4.is_time_range_split())
        # 4PM - 12AM => not split
        d_5 = DailySchedule(weekday=0, start_time=datetime.time(16, 0, 0))
        self.assertFalse(d_5.is_time_range_split())
        # 4AM - 4PM => not split
        d_6 = DailySchedule(weekday=0, start_time=datetime.time(4, 0, 0), end_time=datetime.time(16, 0, 0))
        self.assertFalse(d_6.is_time_range_split())

    def test_earliest_match(self):
        tz = timezone.get_current_timezone()
        start_time_1 = make_aware(datetime.datetime(year=2022, month=2, day=20, hour=17), tz)  # Sun Feb 20 2022 at 5PM
        end_time_1 = make_aware(datetime.datetime(year=2022, month=2, day=23, hour=17), tz)  # Wed Feb 23 2022 at 5PM
        rate_time_1 = RateTime.objects.create(name="rate time 1")
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(16, 0, 0), end_time=datetime.time(20, 0, 0), weekday=0
        )
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(20, 0, 0), end_time=datetime.time(0, 0, 0), weekday=0
        )
        start_result_1 = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=16), tz)  # Mon Feb 21 at 4PM
        end_result_1 = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=20), tz)  # Mon Feb 21 at 8PM
        first_start_1, first_end_1 = rate_time_1.earliest_match(start_time_1, end_time_1)
        self.assertEqual(first_start_1, start_result_1)
        self.assertEqual(first_end_1, end_result_1)

        start_time_2 = datetime.datetime(year=2022, month=2, day=22, hour=17)  # Tue Feb 22 2022 at 5PM
        end_time_2 = datetime.datetime(year=2022, month=2, day=23, hour=17)  # Wed Feb 23 2022 at 5PM
        # no match
        first_start_2, first_end_2 = rate_time_1.earliest_match(start_time_2, end_time_2)
        self.assertIsNone(first_start_2)
        self.assertIsNone(first_end_2)

        # Should work without timezone too (it will assume it's the local)
        start_time_3 = datetime.datetime(year=2022, month=2, day=22, hour=17)  # Tue Feb 22 2022 at 5PM
        end_time_3 = datetime.datetime(year=2022, month=2, day=28, hour=22)  # Mon Feb 28 2022 at 10PM
        start_result_3 = datetime.datetime(year=2022, month=2, day=28, hour=16)  # Mon Feb 28 at 4PM
        end_result_3 = datetime.datetime(year=2022, month=2, day=28, hour=20)  # Mon Feb 28 at 8PM
        first_start_3, first_end_3 = rate_time_1.earliest_match(start_time_3, end_time_3)
        self.assertEqual(first_start_3, start_result_3.replace(tzinfo=tz))
        self.assertEqual(first_end_3, end_result_3.replace(tzinfo=tz))

        start_time_4 = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        end_time_4 = make_aware(datetime.datetime(year=2022, month=2, day=28, hour=22), tz)  # Mon Feb 28 2022 at 10PM
        start_result_4 = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 at 5PM
        end_result_4 = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=20), tz)  # Mon Feb 21 at 8PM
        first_start_4, first_end_4 = rate_time_1.earliest_match(start_time_4, end_time_4)
        self.assertEqual(first_start_4, start_result_4.replace(tzinfo=tz))
        self.assertEqual(first_end_4, end_result_4.replace(tzinfo=tz))

    def test_rate_times_split_previous_day(self):
        # Original: Monday 8PM-4AM (split time)
        rate_time = RateTime.objects.create(name="original split")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(4, 0, 0), weekday=0
        )
        # Sunday 11PM-9PM should overlap (previous day split, stops after start)
        rate_time_1 = RateTime.objects.create(name="previous split 1")
        # We cannot have split Sunday
        with self.assertRaises(ValidationError) as cm:
            DailySchedule(
                rate_time=rate_time_1, start_time=datetime.time(23, 0, 0), end_time=datetime.time(21, 0, 0), weekday=6
            ).full_clean()
        # We cannot have split Sunday
        with self.assertRaises(ValidationError) as cm:
            DailySchedule(
                rate_time=rate_time_1, start_time=datetime.time(23, 0, 0), end_time=datetime.time(23, 0, 0), weekday=6
            ).full_clean()
        # Create it in 2 steps
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(23, 0, 0), end_time=datetime.time(0, 0, 0), weekday=6
        )
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(0, 0, 0), end_time=datetime.time(21, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_1))

        # Sunday 11PM-5PM should NOT overlap (previous day split, stops before start)
        rate_time_2 = RateTime.objects.create(name="previous split 2")
        DailySchedule.objects.create(
            rate_time=rate_time_2, start_time=datetime.time(23, 0, 0), end_time=datetime.time(0, 0, 0), weekday=6
        )
        DailySchedule.objects.create(
            rate_time=rate_time_2, start_time=datetime.time(0, 0, 0), end_time=datetime.time(17, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_2))

        # Sunday 5PM-11PM should NOT overlap (previous day not split)
        rate_time_3 = RateTime.objects.create(name="previous not split 3")
        DailySchedule.objects.create(
            rate_time=rate_time_3, start_time=datetime.time(17, 0, 0), end_time=datetime.time(23, 0, 0), weekday=6
        )
        self.assertFalse(rate_time.overlaps(rate_time_3))

    def test_rate_times_split_same_day(self):
        # Original: Monday 8PM-4AM (split time)
        rate_time = RateTime.objects.create(name="original split")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(4, 0, 0), weekday=0
        )
        # Monday 10PM-11PM should overlap (not split, begins later than start)
        rate_time_1 = RateTime.objects.create(name="split 1")
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(22, 0, 0), end_time=datetime.time(23, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_1))
        # Monday 5PM-5AM should overlap (split, begins before start and stops after end)
        rate_time_2 = RateTime.objects.create(name="split 2")
        DailySchedule.objects.create(
            rate_time=rate_time_2, start_time=datetime.time(17, 0, 0), end_time=datetime.time(5, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_2))
        # Monday 5PM-2AM should overlap (split, begins before start and stops before end)
        rate_time_3 = RateTime.objects.create(name="split 3")
        DailySchedule.objects.create(
            rate_time=rate_time_3, start_time=datetime.time(17, 0, 0), end_time=datetime.time(2, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_3))
        # Monday 10PM-2AM should overlap (split, begins after start and stops before end)
        rate_time_4 = RateTime.objects.create(name="split 4")
        DailySchedule.objects.create(
            rate_time=rate_time_4, start_time=datetime.time(22, 0, 0), end_time=datetime.time(2, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_4))
        # Monday 10PM-5AM should overlap (split begins after start and stops after end)
        rate_time_5 = RateTime.objects.create(name="split 5")
        DailySchedule.objects.create(
            rate_time=rate_time_5, start_time=datetime.time(22, 0, 0), end_time=datetime.time(5, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_5))
        # Monday 1AM-3AM should NOT overlap (not split, begins before start and ends before start)
        rate_time_6 = RateTime.objects.create(name="same 1")
        DailySchedule.objects.create(
            rate_time=rate_time_6, start_time=datetime.time(1, 0, 0), end_time=datetime.time(3, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_6))
        # Monday 1AM-5PM should NOT overlap (not split, begins before start and ends before start)
        rate_time_7 = RateTime.objects.create(name="same 2")
        DailySchedule.objects.create(
            rate_time=rate_time_7, start_time=datetime.time(1, 0, 0), end_time=datetime.time(5, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_7))
        # Monday 4PM-5PM should NOT overlap (not split, begins before start and ends before start)
        rate_time_8 = RateTime.objects.create(name="same 3")
        DailySchedule.objects.create(
            rate_time=rate_time_8, start_time=datetime.time(16, 0, 0), end_time=datetime.time(17, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_8))

    def test_rate_times_split_next_day(self):
        # Original: Monday 8PM-4AM (split time)
        rate_time = RateTime.objects.create(name="original split")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(4, 0, 0), weekday=0
        )
        # Tuesday 1-5AM should overlap (next day, begins before end)
        rate_time_1 = RateTime.objects.create(name="next 1")
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(1, 0, 0), end_time=datetime.time(5, 0, 0), weekday=1
        )
        self.assertTrue(rate_time.overlaps(rate_time_1))
        # Tuesday 1-2AM should overlap (next day, begins & stops before end)
        rate_time_2 = RateTime.objects.create(name="next 2")
        DailySchedule.objects.create(
            rate_time=rate_time_2, start_time=datetime.time(1, 0, 0), end_time=datetime.time(2, 0, 0), weekday=1
        )
        self.assertTrue(rate_time.overlaps(rate_time_2))
        # Tuesday 5-8PM should NOT overlap (next day split, stops before start)
        rate_time_3 = RateTime.objects.create(name="next 3")
        DailySchedule.objects.create(
            rate_time=rate_time_3, start_time=datetime.time(17, 0, 0), end_time=datetime.time(20, 0, 0), weekday=1
        )
        self.assertFalse(rate_time.overlaps(rate_time_3))
        # Tuesday 11PM-5PM should NOT overlap (next day not split)
        rate_time_4 = RateTime.objects.create(name="next split")
        DailySchedule.objects.create(
            rate_time=rate_time_4, start_time=datetime.time(23, 0, 0), end_time=datetime.time(17, 0, 0), weekday=1
        )
        self.assertFalse(rate_time.overlaps(rate_time_4))

    def test_rate_times_previous_day(self):
        # Original: monday 4PM-8PM
        rate_time = RateTime.objects.create(name="original")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(16, 0, 0), end_time=datetime.time(20, 0, 0), weekday=0
        )
        # Sunday 11PM-9PM should overlap (previous day split, stops after start)
        rate_time_1 = RateTime.objects.create(name="previous split 1")
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(23, 0, 0), end_time=datetime.time(0, 0, 0), weekday=6
        )
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(0, 0, 0), end_time=datetime.time(21, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_1))
        # Sunday 11PM-2PM should NOT overlap (previous day split, stops before start)
        rate_time_2 = RateTime.objects.create(name="previous split 2")
        DailySchedule.objects.create(
            rate_time=rate_time_2, start_time=datetime.time(23, 0, 0), end_time=datetime.time(0, 0, 0), weekday=6
        )
        DailySchedule.objects.create(
            rate_time=rate_time_2, start_time=datetime.time(0, 0, 0), end_time=datetime.time(14, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_2))

    def test_rate_times_same_day(self):
        # Original: Monday 4PM-8PM
        rate_time = RateTime.objects.create(name="original")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(16, 0, 0), end_time=datetime.time(20, 0, 0), weekday=0
        )
        # Monday 6AM-2PM should NOT overlap (not split, stops before start)
        rate_time_1 = RateTime.objects.create(name="same 1")
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(6, 0, 0), end_time=datetime.time(14, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_1))
        # Monday 9PM-11PM should NOT overlap (not split, begins after end)
        rate_time_2 = RateTime.objects.create(name="same 2")
        DailySchedule.objects.create(
            rate_time=rate_time_2, start_time=datetime.time(21, 0, 0), end_time=datetime.time(23, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_2))
        # Monday 6AM-11PM should overlap (not split, begins before start and stops after end)
        rate_time_3 = RateTime.objects.create(name="same 3")
        DailySchedule.objects.create(
            rate_time=rate_time_3, start_time=datetime.time(6, 0, 0), end_time=datetime.time(23, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_3))
        # Monday 5PM-6PM should overlap (not split, begins after start and stops before end)
        rate_time_4 = RateTime.objects.create(name="same 4")
        DailySchedule.objects.create(
            rate_time=rate_time_4, start_time=datetime.time(17, 0, 0), end_time=datetime.time(18, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_4))
        # Monday 6AM-6PM should overlap (not split, begins before start and stops before end)
        rate_time_5 = RateTime.objects.create(name="same 5")
        DailySchedule.objects.create(
            rate_time=rate_time_5, start_time=datetime.time(6, 0, 0), end_time=datetime.time(18, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_5))
        # Monday 6PM-11PM should overlap (not split, begins before end and stops after end)
        rate_time_6 = RateTime.objects.create(name="same 6")
        DailySchedule.objects.create(
            rate_time=rate_time_6, start_time=datetime.time(6, 0, 0), end_time=datetime.time(23, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_6))
        # Monday 2PM-2AM should overlap (split, begins before start and stops after end)
        rate_time_7 = RateTime.objects.create(name="split 1")
        DailySchedule.objects.create(
            rate_time=rate_time_7, start_time=datetime.time(18, 0, 0), end_time=datetime.time(2, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_7))
        # Monday 6PM-2AM should overlap (split, begins before end and stops after end)
        rate_time_8 = RateTime.objects.create(name="split 2")
        DailySchedule.objects.create(
            rate_time=rate_time_8, start_time=datetime.time(14, 0, 0), end_time=datetime.time(2, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_8))
        # Monday 11PM-2AM should NOT overlap (split, begins after end)
        rate_time_9 = RateTime.objects.create(name="split 3")
        DailySchedule.objects.create(
            rate_time=rate_time_9, start_time=datetime.time(23, 0, 0), end_time=datetime.time(2, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_9))
        # Monday 11PM-7PM should NOT overlap (split, begins after end)
        rate_time_10 = RateTime.objects.create(name="split 4")
        DailySchedule.objects.create(
            rate_time=rate_time_10, start_time=datetime.time(23, 0, 0), end_time=datetime.time(19, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_10))

    def test_rate_times_edge_cases(self):
        # Original: Monday 4PM-8PM
        rate_time = RateTime.objects.create(name="original")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(16, 0, 0), end_time=datetime.time(20, 0, 0), weekday=0
        )
        # Monday 4PM-8PM should overlap
        rate_time_1 = RateTime.objects.create(name="edge 1")
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(16, 0, 0), end_time=datetime.time(20, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_1))
        # Monday 6AM-4PM should NOT overlap
        rate_time_2 = RateTime.objects.create(name="edge 2")
        DailySchedule.objects.create(
            rate_time=rate_time_2, start_time=datetime.time(6, 0, 0), end_time=datetime.time(16, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_2))
        # Monday 8PM-11PM should NOT overlap
        rate_time_3 = RateTime.objects.create(name="edge 3")
        DailySchedule.objects.create(
            rate_time=rate_time_3, start_time=datetime.time(20, 0, 0), end_time=datetime.time(23, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_3))
        # Monday 8PM-2AM should NOT overlap
        rate_time_4 = RateTime.objects.create(name="edge 4")
        DailySchedule.objects.create(
            rate_time=rate_time_4, start_time=datetime.time(20, 0, 0), end_time=datetime.time(2, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_4))
        # Monday 8PM-2AM should NOT overlap
        rate_time_5 = RateTime.objects.create(name="edge 5")
        DailySchedule.objects.create(
            rate_time=rate_time_5, start_time=datetime.time(20, 0, 0), end_time=datetime.time(2, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_5))
        # Sunday 11PM-4PM should NOT overlap
        rate_time_6 = RateTime.objects.create(name="edge 6")
        DailySchedule.objects.create(
            rate_time=rate_time_6, start_time=datetime.time(23, 0, 0), end_time=datetime.time(0, 0, 0), weekday=6
        )
        DailySchedule.objects.create(
            rate_time=rate_time_6, start_time=datetime.time(0, 0, 0), end_time=datetime.time(16, 0, 0), weekday=0
        )
        self.assertFalse(rate_time.overlaps(rate_time_6))

    def test_rate_times_split_edge_cases(self):
        # Original: Monday 8PM-4AM (split time)
        rate_time = RateTime.objects.create(name="original split")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(4, 0, 0), weekday=0
        )
        # Monday 8PM-4AM should overlap
        rate_time_1 = RateTime.objects.create(name="split 1")
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(20, 0, 0), end_time=datetime.time(4, 0, 0), weekday=0
        )
        self.assertTrue(rate_time.overlaps(rate_time_1))



from typing import Iterable

from NEMO.templatetags.custom_tags_and_filters import app_installed
from django import template

from NEMO_billing.invoices.utilities import display_amount as amount

register = template.Library()


@register.filter
def display_amount(value, configuration=None):
    if value is not None and configuration:
        return amount(value, configuration)


@register.filter
def to_dict(value: Iterable, attribute=None):
    return {getattr(item, attribute, None): item for item in value}


@register.simple_tag()
def cap_discount_installed():
    # CAP discount needs to be installed and the data processor has to be set to CAPDiscountInvoiceDataProcessor
    if app_installed("NEMO_billing.cap_discount"):
        from NEMO_billing.invoices.processors import invoice_data_processor_class as data_processor
        from NEMO_billing.cap_discount.processors import CAPDiscountInvoiceDataProcessor

        return isinstance(data_processor, CAPDiscountInvoiceDataProcessor)
    return False



import datetime
from typing import Dict, List

from NEMO_billing.invoices.models import InvoiceConfiguration

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo

from NEMO.models import AreaAccessRecord, UsageEvent
from django.test import TestCase
from django.utils import timezone
from django.utils.timezone import make_aware

from NEMO_billing.invoices.exceptions import NoRateSetException
from NEMO_billing.invoices.processors import BillableItem, break_up_and_add, find_rates, split_records_by_day
from NEMO_billing.rates.models import DailySchedule, Rate, RateTime, RateType
from NEMO_billing.tests.test_utilities import basic_data


class TestRateTimeSplitBillables(TestCase):
    def setUp(self):
        self.config = InvoiceConfiguration.first_or_default()

    def test_split_billable_items(self):
        test_user, test_project, tool, area = basic_data()
        tz = timezone.get_current_timezone()
        rate_time = RateTime.objects.create(name="rate time")
        # Rate time MON-TUE 4PM-8PM
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(16, 0, 0), end_time=datetime.time(20, 0, 0), weekday=0
        )
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(16, 0, 0), end_time=datetime.time(20, 0, 0), weekday=1
        )
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=10)
        start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        billable = BillableItem(
            UsageEvent(tool=tool, start=start, user=test_user, operator=test_user, end=end, project=test_project),
            test_project,
        )
        billables: List[BillableItem] = []
        # Should fail because we don't have a base rate set
        with self.assertRaises(NoRateSetException):
            find_rates(
                billable.rate_type,
                billable.project,
                self.config,
                billable.tool,
                billable.area,
                billable.consumable,
            )
        base_rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=5)
        # Now should work after we reinit the config which contains all rates
        self.config.init_rate_help()
        rates = find_rates(
            billable.rate_type, billable.project, self.config, billable.tool, billable.area, billable.consumable
        )
        break_up_and_add(billables, rates, billable.start, billable.end, billable)
        # We should have 4 chunks
        self.assertEqual(len(billables), 4)
        first = billables[0]
        second = billables[1]
        third = billables[2]
        last = billables[3]
        # First chunk: start until Monday, Feb 21 at 8PM
        self.assertEqual(first.start, start)
        self.assertEqual(first.end, make_aware(datetime.datetime(year=2022, month=2, day=21, hour=20), tz))
        self.assertEqual(first.rate, rate_with_time)
        self.assertEqual(first.project, test_project)
        # Second chunk: start Monday, Feb 21 at 8PM until Tuesday, Feb 22 4PM, base rate
        self.assertEqual(second.start, make_aware(datetime.datetime(year=2022, month=2, day=21, hour=20), tz))
        self.assertEqual(second.end, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=16), tz))
        self.assertEqual(second.rate, base_rate)
        self.assertEqual(second.project, test_project)
        # Third chunk: start Tuesday, Feb 22 at 4PM until Tuesday, Feb 22 8PM
        self.assertEqual(third.start, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=16), tz))
        self.assertEqual(third.end, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=20), tz))
        self.assertEqual(third.rate, rate_with_time)
        self.assertEqual(third.project, test_project)
        # Last chunk: start Tuesday, Feb 22 at 8PM until end
        self.assertEqual(last.start, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=20), tz))
        self.assertEqual(last.end, end)
        self.assertEqual(last.rate, base_rate)
        self.assertEqual(last.project, test_project)

    def test_split_billable_items_no_times(self):
        test_user, test_project, tool, area = basic_data()
        tz = timezone.get_current_timezone()
        rate_time = RateTime.objects.create(name="rate time")
        # Rate time MON-WED full days
        DailySchedule.objects.create(rate_time=rate_time, weekday=0)
        DailySchedule.objects.create(rate_time=rate_time, weekday=2)
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=10)
        start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        billable = BillableItem(
            UsageEvent(tool=tool, start=start, user=test_user, operator=test_user, end=end, project=test_project),
            test_project,
        )
        billables: List[BillableItem] = []
        # Should fail because we don't have a base rate set
        with self.assertRaises(NoRateSetException):
            find_rates(
                billable.rate_type,
                billable.project,
                self.config,
                billable.tool,
                billable.area,
                billable.consumable,
            )
        base_rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=5)
        # Now should work after we reinit the config which contains all rates
        self.config.init_rate_help()
        rates = find_rates(
            billable.rate_type, billable.project, self.config, billable.tool, billable.area, billable.consumable
        )
        break_up_and_add(billables, rates, billable.start, billable.end, billable)
        # We should have 4 chunks
        self.assertEqual(len(billables), 4)
        first = billables[0]
        second = billables[1]
        third = billables[2]
        last = billables[3]
        # First chunk: start until Tuesday, Feb 22 at 0:00
        self.assertEqual(first.start, start)
        self.assertEqual(first.end, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=0), tz))
        self.assertEqual(first.rate, rate_with_time)
        self.assertEqual(first.project, test_project)
        # Second chunk: start Tuesday, Feb 22 at 0:00 until Wednesday, Feb 23 0:00, base rate
        self.assertEqual(second.start, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=0), tz))
        self.assertEqual(second.end, make_aware(datetime.datetime(year=2022, month=2, day=23, hour=0), tz))
        self.assertEqual(second.rate, base_rate)
        self.assertEqual(second.project, test_project)
        # Third chunk: start Wednesday, Feb 23 at 0:00 until Thursday, Feb 24 0:00
        self.assertEqual(third.start, make_aware(datetime.datetime(year=2022, month=2, day=23, hour=0), tz))
        self.assertEqual(third.end, make_aware(datetime.datetime(year=2022, month=2, day=24, hour=0), tz))
        self.assertEqual(third.rate, rate_with_time)
        self.assertEqual(third.project, test_project)
        # Last chunk: start Tuesday, Feb 22 at 8PM until end
        self.assertEqual(last.start, make_aware(datetime.datetime(year=2022, month=2, day=24, hour=0), tz))
        self.assertEqual(last.end, end)
        self.assertEqual(last.rate, base_rate)
        self.assertEqual(last.project, test_project)

    def test_split_billable_items_split_rate_time(self):
        test_user, test_project, tool, area = basic_data()
        tz = timezone.get_current_timezone()
        # Rate time MON-TUE 8PM-10AM
        rate_time = RateTime.objects.create(name="rate time")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(10, 0, 0), weekday=0
        )
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(10, 0, 0), weekday=1
        )
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=10)
        start = make_aware(datetime.datetime(year=2022, month=2, day=22, hour=4), tz)  # Tue Feb 22 2022 at 4AM
        end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        billable = BillableItem(
            UsageEvent(tool=tool, start=start, user=test_user, operator=test_user, end=end, project=test_project),
            test_project,
        )
        billables: List[BillableItem] = []
        base_rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=5)
        rates = find_rates(
            billable.rate_type, billable.project, self.config, billable.tool, billable.area, billable.consumable
        )
        break_up_and_add(billables, rates, billable.start, billable.end, billable)
        # We should have 4 chunks
        self.assertEqual(len(billables), 4)
        first = billables[0]
        second = billables[1]
        third = billables[2]
        last = billables[3]
        # First chunk: start until TUE, Feb 22 at 10AM
        self.assertEqual(first.start, start)
        self.assertEqual(first.end, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=10), tz))
        self.assertEqual(first.rate, rate_with_time)
        self.assertEqual(first.project, test_project)
        # Second chunk: start TUE, Feb 22 at 10AM until TUE, Feb 22 8PM, base rate
        self.assertEqual(second.start, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=10), tz))
        self.assertEqual(second.end, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=20), tz))
        self.assertEqual(second.rate, base_rate)
        self.assertEqual(second.project, test_project)
        # Third chunk: start TUE, Feb 22 at 8PM until WED, Feb 23 10AM
        self.assertEqual(third.start, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=20), tz))
        self.assertEqual(third.end, make_aware(datetime.datetime(year=2022, month=2, day=23, hour=10), tz))
        self.assertEqual(third.rate, rate_with_time)
        self.assertEqual(third.project, test_project)
        # Last chunk: start WED, Feb 23 at 10AM until end
        self.assertEqual(last.start, make_aware(datetime.datetime(year=2022, month=2, day=23, hour=10), tz))
        self.assertEqual(last.end, end)
        self.assertEqual(last.rate, base_rate)
        self.assertEqual(last.project, test_project)

    def test_split_billable_items_split_rate_time_end(self):
        test_user, test_project, tool, area = basic_data()
        tz = timezone.get_current_timezone()
        # Rate time TUE 8PM-10AM
        rate_time = RateTime.objects.create(name="rate time")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(10, 0, 0), weekday=1
        )
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=10)
        start = make_aware(datetime.datetime(year=2022, month=2, day=22, hour=4), tz)  # TUE Feb 22 2022 at 4AM
        end = make_aware(datetime.datetime(year=2022, month=2, day=22, hour=21), tz)  # TUE Feb 22 2022 at 9PM
        billable = BillableItem(
            UsageEvent(tool=tool, start=start, user=test_user, operator=test_user, end=end, project=test_project),
            test_project,
        )
        billables: List[BillableItem] = []
        base_rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=5)
        rates = find_rates(
            billable.rate_type, billable.project, self.config, billable.tool, billable.area, billable.consumable
        )
        break_up_and_add(billables, rates, billable.start, billable.end, billable)
        # We should have 2 chunks
        self.assertEqual(len(billables), 2)
        first = billables[0]
        last = billables[1]
        # First chunk: start until TUE, Feb 22 at 8PM
        self.assertEqual(first.start, start)
        self.assertEqual(first.end, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=20), tz))
        self.assertEqual(first.rate, base_rate)
        self.assertEqual(first.project, test_project)
        # Last chunk: start TUE, Feb 22 at 8PM until end
        self.assertEqual(last.start, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=20), tz))
        self.assertEqual(last.end, end)
        self.assertEqual(last.rate, rate_with_time)
        self.assertEqual(last.project, test_project)

    def test_split_billable_items_no_overlap(self):
        test_user, test_project, tool, area = basic_data()
        tz = timezone.get_current_timezone()
        # Rate time MON 8PM-10AM
        rate_time = RateTime.objects.create(name="rate time")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(10, 0, 0), weekday=0
        )
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=10)
        start = make_aware(datetime.datetime(year=2022, month=2, day=22, hour=13), tz)  # Tue Feb 22 2022 at 1PM
        end = make_aware(datetime.datetime(year=2022, month=2, day=22, hour=15), tz)  # Tue Feb 22 2022 at 3PM
        billable = BillableItem(
            UsageEvent(tool=tool, start=start, user=test_user, operator=test_user, end=end, project=test_project),
            test_project,
        )
        billables: List[BillableItem] = []
        base_rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=5)
        rates = find_rates(
            billable.rate_type, billable.project, self.config, billable.tool, billable.area, billable.consumable
        )
        break_up_and_add(billables, rates, billable.start, billable.end, billable)
        # We should have 1 chunk
        self.assertEqual(len(billables), 1)
        first = billables[0]
        # First and only chunk: start until end
        self.assertEqual(first.start, start)
        self.assertEqual(first.end, end)
        self.assertEqual(first.rate, base_rate)
        self.assertEqual(first.project, test_project)

    def test_split_billable_items_dst_forward(self):
        test_user, test_project, tool, area = basic_data()
        # Test with DST (EST = March 13 2AM -> 3AM)
        tz = zoneinfo.ZoneInfo("America/New_York")
        # Rate time SAT 8PM-2AM
        rate_time = RateTime.objects.create(name="rate time")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(2, 0, 0), weekday=5
        )
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=10)
        start = make_aware(datetime.datetime(year=2022, month=3, day=13, hour=1), tz)  # Sun Mar 13 2022 at 1AM
        end = make_aware(datetime.datetime(year=2022, month=3, day=13, hour=3), tz)  # Sun Mar 13 2022 at 3AM
        billable = BillableItem(
            UsageEvent(tool=tool, start=start, user=test_user, operator=test_user, end=end, project=test_project),
            test_project,
        )
        billables: List[BillableItem] = []
        base_rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=5)
        rates = find_rates(
            billable.rate_type, billable.project, self.config, billable.tool, billable.area, billable.consumable
        )
        break_up_and_add(billables, rates, billable.start, billable.end, billable)
        # We should have 1 chunk
        self.assertEqual(len(billables), 1)
        first = billables[0]
        # First chunk: start until end, but it should be 1 hours, not 2
        self.assertEqual(first.start, start)
        self.assertEqual(first.end, end)
        self.assertEqual(first.rate, rate_with_time)
        self.assertEqual(first.project, test_project)
        self.assertNotEqual(first.quantity, 2 * 60.0)
        self.assertEqual(first.quantity, 1 * 60.0)

    def test_split_billable_items_dst_backwards(self):
        test_user, test_project, tool, area = basic_data()
        # Test with DST (EST = NOV 6 2AM -> 1AM)
        tz = zoneinfo.ZoneInfo("America/New_York")
        # Rate time SAT 8PM-2AM
        rate_time = RateTime.objects.create(name="rate time")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(2, 0, 0), weekday=5
        )
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=10)
        start = make_aware(datetime.datetime(year=2022, month=11, day=6, hour=0), tz)  # Sun NOV 6 2022 at 12AM
        end = make_aware(datetime.datetime(year=2022, month=11, day=6, hour=2), tz)  # Sun NOV 6 2022 at 2AM (new 2AM)
        billable = BillableItem(
            UsageEvent(tool=tool, start=start, user=test_user, operator=test_user, end=end, project=test_project),
            test_project,
        )
        billables: List[BillableItem] = []
        base_rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=5)
        rates = find_rates(
            billable.rate_type, billable.project, self.config, billable.tool, billable.area, billable.consumable
        )
        break_up_and_add(billables, rates, billable.start, billable.end, billable)
        # We should have 2 chunks
        self.assertEqual(len(billables), 2)
        first = billables[0]
        last = billables[1]
        # First chunk: start until 2AM old (1AM new), but it should be 2 hours, not 1
        self.assertEqual(first.start, start)
        self.assertEqual(
            first.end, make_aware(datetime.datetime(year=2022, month=11, day=6, hour=6), datetime.timezone.utc)
        )
        self.assertEqual(first.rate, rate_with_time)
        self.assertEqual(first.project, test_project)
        self.assertEqual(first.quantity, 2 * 60.0)
        self.assertNotEqual(first.quantity, 1 * 60.0)
        # Last chunk: from 2AM old (1AM new) until end but it should be 1 hour, not 0
        # Both chunk will not have the rate time, second part will have the regular rate
        self.assertEqual(
            last.start, make_aware(datetime.datetime(year=2022, month=11, day=6, hour=6), datetime.timezone.utc)
        )
        self.assertEqual(last.end, end)
        self.assertEqual(last.rate, base_rate)
        self.assertEqual(last.project, test_project)
        self.assertEqual(last.quantity, 1 * 60.0)
        self.assertNotEqual(last.quantity, 0 * 60.0)

        start_2 = make_aware(datetime.datetime(year=2022, month=11, day=6, hour=0), tz)  # Sun NOV 6 2022 at 12AM
        end_2 = make_aware(datetime.datetime(year=2022, month=11, day=6, hour=4), tz)  # Sun NOV 6 2022 at 4AM (new 4AM)
        billable_2 = BillableItem(
            UsageEvent(tool=tool, start=start_2, user=test_user, operator=test_user, end=end_2, project=test_project),
            test_project,
        )
        billables_2: List[BillableItem] = []
        break_up_and_add(billables_2, rates, billable_2.start, billable_2.end, billable_2)
        # We should have only 2 chunks again (second part is using base rate)
        self.assertEqual(len(billables_2), 2)
        last_one = billables_2[1]
        # First chunk same as before
        # Last chunk: 1AM until 4AM
        self.assertEqual(
            last_one.start, make_aware(datetime.datetime(year=2022, month=11, day=6, hour=6), datetime.timezone.utc)
        )
        self.assertEqual(last_one.end, end_2)
        self.assertEqual(last_one.rate, base_rate)
        self.assertEqual(last_one.project, test_project)
        self.assertEqual(last_one.quantity, 3 * 60.0)

    def test_split_billable_items_complete_overlap(self):
        test_user, test_project, tool, area = basic_data()
        tz = timezone.get_current_timezone()
        # Rate time MON 8PM-10AM
        rate_time = RateTime.objects.create(name="rate time")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(10, 0, 0), weekday=0
        )
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=10)
        start = make_aware(datetime.datetime(year=2022, month=2, day=22, hour=4), tz)  # Tue Feb 22 2022 at 4AM
        end = make_aware(datetime.datetime(year=2022, month=2, day=22, hour=5), tz)  # Tue Feb 22 2022 at 5AM
        billable = BillableItem(
            UsageEvent(tool=tool, start=start, user=test_user, operator=test_user, end=end, project=test_project),
            test_project,
        )
        billables: List[BillableItem] = []
        base_rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=5)
        rates = find_rates(
            billable.rate_type, billable.project, self.config, billable.tool, billable.area, billable.consumable
        )
        break_up_and_add(billables, rates, billable.start, billable.end, billable)
        # We should have 1 chunk
        self.assertEqual(len(billables), 1)
        first = billables[0]
        # First and only chunk: start until TUE, Feb 22 at 5AM
        self.assertEqual(first.start, start)
        self.assertEqual(first.end, make_aware(datetime.datetime(year=2022, month=2, day=22, hour=5), tz))
        self.assertEqual(first.rate, rate_with_time)
        self.assertEqual(first.project, test_project)

    def test_split_billable_items_max_time_in_between(self):
        test_user, test_project, tool, area = basic_data()
        tz = timezone.get_current_timezone()
        # Rate time MON 8PM-10AM
        rate_time = RateTime.objects.create(name="rate time")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(10, 0, 0), weekday=0
        )
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=10)
        start = make_aware(datetime.datetime(year=2022, month=2, day=22, hour=11), tz)  # TUE Feb 22 2022 at 11AM
        end = make_aware(datetime.datetime(year=2022, month=3, day=1, hour=16), tz)  # TUE Mar 1 2022 at 4PM
        billable = BillableItem(
            UsageEvent(tool=tool, start=start, user=test_user, operator=test_user, end=end, project=test_project),
            test_project,
        )
        billables: List[BillableItem] = []
        base_rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=5)
        rates = find_rates(
            billable.rate_type, billable.project, self.config, billable.tool, billable.area, billable.consumable
        )
        break_up_and_add(billables, rates, billable.start, billable.end, billable)
        # We should have 3 chunks
        self.assertEqual(len(billables), 3)
        first = billables[0]
        second = billables[1]
        last = billables[2]
        # First chunk: start until MON, Feb 28 at 8PM
        self.assertEqual(first.start, start)
        self.assertEqual(first.end, make_aware(datetime.datetime(year=2022, month=2, day=28, hour=20), tz))
        self.assertEqual(first.rate, base_rate)
        self.assertEqual(first.project, test_project)
        # Second chunk: MON, Feb 28 at 8PM to TUE, Feb 29 at 10AM
        self.assertEqual(second.start, make_aware(datetime.datetime(year=2022, month=2, day=28, hour=20), tz))
        self.assertEqual(second.end, make_aware(datetime.datetime(year=2022, month=3, day=1, hour=10), tz))
        self.assertEqual(second.rate, rate_with_time)
        self.assertEqual(second.project, test_project)
        # last chunk: TUE, Feb 29 at 10AM until end
        self.assertEqual(last.start, make_aware(datetime.datetime(year=2022, month=3, day=1, hour=10), tz))
        self.assertEqual(last.end, end)
        self.assertEqual(last.rate, base_rate)
        self.assertEqual(last.project, test_project)

    def test_split_billable_items_multiple_rate_times(self):
        test_user, test_project, tool, area = basic_data()
        tz = timezone.get_current_timezone()
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        # Rate time MON 8PM-10AM
        rate_time = RateTime.objects.create(name="rate time")
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(20, 0, 0), end_time=datetime.time(10, 0, 0), weekday=0
        )
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=10)
        # Rate time WED 10PM-11PM
        rate_time_2 = RateTime.objects.create(name="rate time 2")
        DailySchedule.objects.create(
            rate_time=rate_time_2, start_time=datetime.time(22, 0, 0), end_time=datetime.time(23, 0, 0), weekday=2
        )
        rate_with_time_2 = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time_2, amount=25)
        start = make_aware(datetime.datetime(year=2022, month=2, day=22, hour=11), tz)  # TUE Feb 22 2022 at 11AM
        end = make_aware(datetime.datetime(year=2022, month=3, day=1, hour=16), tz)  # TUE Mar 1 2022 at 4PM
        billable = BillableItem(
            UsageEvent(tool=tool, start=start, user=test_user, operator=test_user, end=end, project=test_project),
            test_project,
        )
        billables: List[BillableItem] = []
        base_rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=5)
        rates = find_rates(
            billable.rate_type, billable.project, self.config, billable.tool, billable.area, billable.consumable
        )
        break_up_and_add(billables, rates, billable.start, billable.end, billable)
        # We should have 5 chunks
        self.assertEqual(len(billables), 5)
        first = billables[0]
        second = billables[1]
        third = billables[2]
        fourth = billables[3]
        last = billables[4]
        # First chunk: start until WED, Feb 23 at 10PM
        self.assertEqual(first.start, start)
        self.assertEqual(first.end, make_aware(datetime.datetime(year=2022, month=2, day=23, hour=22), tz))
        self.assertEqual(first.rate, base_rate)
        self.assertEqual(first.project, test_project)
        # Second chunk:  WED, Feb 23 at 10PM until WED, Feb 23 at 11PM
        self.assertEqual(second.start, make_aware(datetime.datetime(year=2022, month=2, day=23, hour=22), tz))
        self.assertEqual(second.end, make_aware(datetime.datetime(year=2022, month=2, day=23, hour=23), tz))
        self.assertEqual(second.rate, rate_with_time_2)
        self.assertEqual(second.project, test_project)
        # Third chunk: WED, Feb 23 at 11PM until MON, Feb 28 at 8PM
        self.assertEqual(third.start, make_aware(datetime.datetime(year=2022, month=2, day=23, hour=23), tz))
        self.assertEqual(third.end, make_aware(datetime.datetime(year=2022, month=2, day=28, hour=20), tz))
        self.assertEqual(third.rate, base_rate)
        self.assertEqual(third.project, test_project)
        # Fourth chunk: MON, Feb 28 at 8PM to TUE, Feb 29 at 10AM
        self.assertEqual(fourth.start, make_aware(datetime.datetime(year=2022, month=2, day=28, hour=20), tz))
        self.assertEqual(fourth.end, make_aware(datetime.datetime(year=2022, month=3, day=1, hour=10), tz))
        self.assertEqual(fourth.rate, rate_with_time)
        self.assertEqual(fourth.project, test_project)
        # last chunk: TUE, Feb 29 at 10AM until end
        self.assertEqual(last.start, make_aware(datetime.datetime(year=2022, month=3, day=1, hour=10), tz))
        self.assertEqual(last.end, end)
        self.assertEqual(last.rate, base_rate)
        self.assertEqual(last.project, test_project)

    def test_split_area_access_records(self):
        test_user, test_project, tool, area = basic_data()
        now = make_aware(datetime.datetime.now())
        record = AreaAccessRecord.objects.create(
            customer=test_user,
            project=test_project,
            area=area,
            start=now - datetime.timedelta(minutes=1),
            end=now,
        )
        # Should be unchanged
        self.assertEqual([record], split_records_by_day(record))
        # Now let's add one overnight
        record = AreaAccessRecord.objects.create(
            customer=test_user,
            project=test_project,
            area=area,
            start=now - datetime.timedelta(hours=23, minutes=59),
            end=now,
        )
        new_records = split_records_by_day(record)
        midnight_today = now.astimezone().replace(hour=0, minute=0, second=0, microsecond=0)
        self.assertEqual(2, len(new_records))
        self.assertEqual(new_records[0].area, record.area)
        self.assertEqual(new_records[0].customer, record.customer)
        self.assertEqual(new_records[0].project, record.project)
        self.assertEqual(new_records[0].start, record.start)
        self.assertEqual(new_records[0].end, midnight_today)
        self.assertEqual(new_records[1].area, record.area)
        self.assertEqual(new_records[1].customer, record.customer)
        self.assertEqual(new_records[1].project, record.project)
        self.assertEqual(new_records[1].start, midnight_today)
        self.assertEqual(new_records[1].end, record.end)
        # Have it span across more than one day
        record_start = now - datetime.timedelta(days=7)
        record = AreaAccessRecord.objects.create(
            customer=test_user,
            project=test_project,
            area=area,
            start=record_start,
            end=now,
        )
        # request them now, should have 8
        new_records = split_records_by_day(record)
        self.assertEqual(8, len(new_records))
        self.assertEqual(new_records[0].start, record.start)
        self.assertEqual(new_records[7].end, record.end)


def get_area_dict(area_access_record) -> Dict:
    return {key: value for key, value in area_access_record.__dict__.items() if key != "_state"}
